#!/usr/bin/env python
# Copyright 2014-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#         Timothy Berkelbach <tim.berkelbach@gmail.com>
#

import os
import sys
import json
import ctypes
import warnings
import numpy as np
import scipy.linalg
from scipy.special import erf, erfc
import pyscf.lib.parameters as param
from pyscf import lib
from pyscf.dft import radi
from pyscf.lib import logger
from pyscf.gto import mole
from pyscf.gto import moleintor
from pyscf.gto.mole import conc_env, is_au # noqa
from pyscf.pbc.gto import _pbcintor
from pyscf.pbc.gto.eval_gto import eval_gto as pbc_eval_gto
from pyscf.pbc.tools import pbc as pbctools
from pyscf import __config__

INTEGRAL_PRECISION = getattr(__config__, 'pbc_gto_cell_Cell_precision', 1e-8)
WRAP_AROUND = getattr(__config__, 'pbc_gto_cell_make_kpts_wrap_around', False)
WITH_GAMMA = getattr(__config__, 'pbc_gto_cell_make_kpts_with_gamma', True)
EXP_DELIMITER = getattr(__config__, 'pbc_gto_cell_split_basis_exp_delimiter',
                        [1.0, 0.5, 0.25, 0.1, 0])
# defined in lib/pbc/cell.h
RCUT_EPS = 1e-3
RCUT_MAX_CYCLE = 10

libpbc = _pbcintor.libpbc

def M(*args, **kwargs):
    r'''This is a shortcut to build up Cell object.

    Examples:

    >>> from pyscf.pbc import gto
    >>> cell = gto.M(a=numpy.eye(3)*4, atom='He 1 1 1', basis='6-31g')
    '''
    cell = Cell()
    cell.build(*args, **kwargs)
    return cell
C = M

def pack(cell):
    '''Pack the input args of :class:`Cell` to a dict, which can be serialized
    with :mod:`pickle`
    '''
    cldic = mole.pack(cell)
    cldic['a'] = cell.a
    cldic['fractional'] = cell.fractional
    cldic['precision'] = cell.precision
    cldic['ke_cutoff'] = cell.ke_cutoff
    cldic['exp_to_discard'] = cell.exp_to_discard
    cldic['_mesh'] = cell._mesh
    cldic['_rcut'] = cell._rcut
    cldic['dimension'] = cell.dimension
    cldic['low_dim_ft_type'] = cell.low_dim_ft_type
    return cldic

def unpack(celldic):
    '''Convert the packed dict to a :class:`Cell` object, to generate the
    input arguments for :class:`Cell` object.
    '''
    cl = Cell()
    cl.__dict__.update(celldic)
    return cl


def dumps(cell):
    '''Serialize Cell object to a JSON formatted str.
    '''
    exclude_keys = {'output', 'stdout', '_keys', '_ctx_lock',
                    'symm_orb', 'irrep_id', 'irrep_name', 'lattice_symmetry'}

    celldic = dict(cell.__dict__)
    for k in exclude_keys:
        if k in celldic:
            del (celldic[k])
    for k in celldic:
        if isinstance(celldic[k], (np.ndarray, np.generic)):
            celldic[k] = celldic[k].tolist()
    celldic['atom'] = repr(cell.atom)
    celldic['basis']= repr(cell.basis)
    celldic['pseudo'] = repr(cell.pseudo)
    celldic['ecp'] = repr(cell.ecp)

    try:
        return json.dumps(celldic)
    except TypeError:
        def skip_value(dic):
            dic1 = {}
            for k,v in dic.items():
                if (v is None or
                    isinstance(v, (str, bool, int, float))):
                    dic1[k] = v
                elif isinstance(v, (list, tuple)):
                    dic1[k] = v   # Should I recursively skip_vaule?
                elif isinstance(v, set):
                    dic1[k] = list(v)
                elif isinstance(v, dict):
                    dic1[k] = skip_value(v)
                elif isinstance(v, np.generic):
                    dic1[k] = v.tolist()
                else:
                    msg =('Function cell.dumps drops attribute %s because '
                          'it is not JSON-serializable' % k)
                    warnings.warn(msg)
            return dic1
        return json.dumps(skip_value(celldic), skipkeys=True)

def loads(cellstr):
    '''Deserialize a str containing a JSON document to a Cell object.
    '''
    from numpy import array  # noqa
    celldic = json.loads(cellstr)
    cell = Cell()
    cell.__dict__.update(celldic)
    cell.atom = eval(cell.atom)
    cell.basis = eval(cell.basis)
    cell.pseudo = eval(cell.pseudo)
    cell.ecp = eval(cell.ecp)
    cell._atm = np.array(cell._atm, dtype=np.int32)
    cell._bas = np.array(cell._bas, dtype=np.int32)
    cell._env = np.array(cell._env, dtype=np.double)
    cell._ecpbas = np.array(cell._ecpbas, dtype=np.int32)
    cell._mesh = np.array(cell._mesh)

    # Symmetry class cannot be serialized by dumps function.
    # Recreate it manually
    if cell.natm > 0 and cell.space_group_symmetry:
        cell.build_lattice_symmetry()

    return cell

def conc_cell(cell1, cell2):
    '''Concatenate two Cell objects.
    '''
    mol3 = mole.conc_mol(cell1, cell2)
    cell3 = Cell()
    cell3.__dict__.update(mol3.__dict__)

    # lattice_vectors needs to be consistent with cell3.unit (Bohr)
    cell3.a = cell1.lattice_vectors()
    cell3.mesh = np.max((cell1.mesh, cell2.mesh), axis=0)

    ke_cutoff1 = cell1.ke_cutoff
    ke_cutoff2 = cell2.ke_cutoff
    if ke_cutoff1 is None and ke_cutoff2 is None:
        cell3.ke_cutoff = None
    else:
        if ke_cutoff1 is None:
            ke_cutoff1 = estimate_ke_cutoff(cell1, cell1.precision)
        if ke_cutoff2 is None:
            ke_cutoff2 = estimate_ke_cutoff(cell2, cell2.precision)
        cell3.ke_cutoff = max(ke_cutoff1, ke_cutoff2)

    cell3.precision = min(cell1.precision, cell2.precision)
    cell3.dimension = max(cell1.dimension, cell2.dimension)
    cell3.low_dim_ft_type = cell1.low_dim_ft_type or cell2.low_dim_ft_type
    cell3.rcut = max(cell1.rcut, cell2.rcut)
    return cell3

def intor_cross(intor, cell1, cell2, comp=None, hermi=0, kpts=None, kpt=None,
                shls_slice=None, **kwargs):
    r'''1-electron integrals from two cells like

    .. math::

        \langle \mu | intor | \nu \rangle, \mu \in cell1, \nu \in cell2
    '''
    intor, comp = moleintor._get_intor_and_comp(cell1._add_suffix(intor), comp)

    if kpts is None:
        if kpt is not None:
            kpts_lst = np.reshape(kpt, (1,3))
        else:
            kpts_lst = np.zeros((1,3))
    else:
        kpts_lst = np.reshape(kpts, (-1,3))
    nkpts = len(kpts_lst)

    pcell = cell1.copy(deep=False)
    pcell.precision = min(cell1.precision, cell2.precision)
    pcell._atm, pcell._bas, pcell._env = \
            atm, bas, env = conc_env(cell1._atm, cell1._bas, cell1._env,
                                     cell2._atm, cell2._bas, cell2._env)
    if shls_slice is None:
        shls_slice = (0, cell1.nbas, 0, cell2.nbas)
    i0, i1, j0, j1 = shls_slice[:4]
    j0 += cell1.nbas
    j1 += cell1.nbas
    ao_loc = moleintor.make_loc(bas, intor)
    ni = ao_loc[i1] - ao_loc[i0]
    nj = ao_loc[j1] - ao_loc[j0]
    out = np.empty((nkpts,comp,ni,nj), dtype=np.complex128)

    if hermi == 0:
        aosym = 's1'
    else:
        aosym = 's2'
    fill = getattr(libpbc, 'PBCnr2c_fill_k'+aosym)
    fintor = getattr(moleintor.libcgto, intor)
    cintopt = lib.c_null_ptr()

    rcut = max(cell1.rcut, cell2.rcut)
    Ls = cell1.get_lattice_Ls(rcut=rcut)
    expkL = np.asarray(np.exp(1j*np.dot(kpts_lst, Ls.T)), order='C')
    drv = libpbc.PBCnr2c_drv

    kderiv = kwargs.get('kderiv', 0)
    if kderiv > 0:
        hermi = 0
        aosym = 's1'
        mat = np.empty((nkpts,(3**kderiv)*comp,ni,nj), dtype=np.complex128)
        if kderiv == 1:
            fac = 1j * lib.einsum('kl,lx->xkl', expkL, Ls)
        elif kderiv == 2:
            fac = -lib.einsum('kl,lx,ly->xykl', expkL, Ls, Ls).reshape(-1,nkpts,len(Ls))
        else:
            raise NotImplementedError

        for x in range(fac.shape[0]):
            facx = np.asarray(fac[x], order='C')
            drv(fintor, fill, out.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(nkpts), ctypes.c_int(comp), ctypes.c_int(len(Ls)),
                Ls.ctypes.data_as(ctypes.c_void_p),
                facx.ctypes.data_as(ctypes.c_void_p),
                (ctypes.c_int*4)(i0, i1, j0, j1),
                ao_loc.ctypes.data_as(ctypes.c_void_p), cintopt,
                atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(pcell.natm),
                bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(pcell.nbas),
                env.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(env.size))

            for k, kpt in enumerate(kpts_lst):
                v = out[k]
                if hermi != 0:
                    for ic in range(comp):
                        lib.hermi_triu(v[ic], hermi=hermi, inplace=True)
                mat[k, x*comp:(x+1)*comp] = v.copy()
        return mat

    drv(fintor, fill, out.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(nkpts), ctypes.c_int(comp), ctypes.c_int(len(Ls)),
        Ls.ctypes.data_as(ctypes.c_void_p),
        expkL.ctypes.data_as(ctypes.c_void_p),
        (ctypes.c_int*4)(i0, i1, j0, j1),
        ao_loc.ctypes.data_as(ctypes.c_void_p), cintopt,
        atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(pcell.natm),
        bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(pcell.nbas),
        env.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(env.size))

    mat = []
    for k, kpt in enumerate(kpts_lst):
        v = out[k]
        if hermi != 0:
            for ic in range(comp):
                lib.hermi_triu(v[ic], hermi=hermi, inplace=True)
        if comp == 1:
            v = v[0]
        if abs(kpt).sum() < 1e-9:  # gamma_point
            v = v.real
        mat.append(v)

    if kpts is None or np.shape(kpts) == (3,):  # A single k-point
        mat = mat[0]
    return mat

def _intor_cross_screened(
        intor, cell1, cell2, comp=None, hermi=0, kpts=None, kpt=None,
        shls_slice=None, **kwargs):
    '''`intor_cross` with prescreening.

    Notes:
         This function may be subject to change.
    '''
    from pyscf.pbc.gto.neighborlist import NeighborListOpt
    intor, comp = moleintor._get_intor_and_comp(cell1._add_suffix(intor), comp)

    if kpts is None:
        if kpt is not None:
            kpts_lst = np.reshape(kpt, (1,3))
        else:
            kpts_lst = np.zeros((1,3))
    else:
        kpts_lst = np.reshape(kpts, (-1,3))
    nkpts = len(kpts_lst)

    pcell = cell1.copy(deep=False)
    pcell.precision = min(cell1.precision, cell2.precision)
    pcell._atm, pcell._bas, pcell._env = \
            atm, bas, env = conc_env(cell1._atm, cell1._bas, cell1._env,
                                     cell2._atm, cell2._bas, cell2._env)
    if shls_slice is None:
        shls_slice = (0, cell1.nbas, 0, cell2.nbas)
    i0, i1, j0, j1 = shls_slice[:4]
    j0 += cell1.nbas
    j1 += cell1.nbas
    ao_loc = moleintor.make_loc(bas, intor)
    ni = ao_loc[i1] - ao_loc[i0]
    nj = ao_loc[j1] - ao_loc[j0]
    out = np.empty((nkpts,comp,ni,nj), dtype=np.complex128)

    if hermi == 0:
        aosym = 's1'
    else:
        aosym = 's2'
    fill = getattr(libpbc, 'PBCnr2c_screened_fill_k'+aosym)
    fintor = getattr(moleintor.libcgto, intor)
    drv = libpbc.PBCnr2c_screened_drv

    rcut = max(cell1.rcut, cell2.rcut)
    Ls = cell1.get_lattice_Ls(rcut=rcut)
    expkL = np.asarray(np.exp(1j*np.dot(kpts_lst, Ls.T)), order='C')

    neighbor_list = kwargs.get('neighbor_list', None)
    if neighbor_list is None:
        nlopt = NeighborListOpt(cell1)
        nlopt.build(cell1, cell2, Ls, set_optimizer=False)
        neighbor_list = nlopt.nl

    cintopt = lib.c_null_ptr()

    drv(fintor, fill, out.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(nkpts), ctypes.c_int(comp), ctypes.c_int(len(Ls)),
        Ls.ctypes.data_as(ctypes.c_void_p),
        expkL.ctypes.data_as(ctypes.c_void_p),
        (ctypes.c_int*4)(i0, i1, j0, j1),
        ao_loc.ctypes.data_as(ctypes.c_void_p), cintopt,
        atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(pcell.natm),
        bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(pcell.nbas),
        env.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(env.size),
        ctypes.byref(neighbor_list))

    nlopt = None

    mat = []
    for k, kpt in enumerate(kpts_lst):
        v = out[k]
        if hermi != 0:
            for ic in range(comp):
                lib.hermi_triu(v[ic], hermi=hermi, inplace=True)
        if comp == 1:
            v = v[0]
        if abs(kpt).sum() < 1e-9:  # gamma_point
            v = v.real
        mat.append(v)

    if kpts is None or np.shape(kpts) == (3,):  # A single k-point
        mat = mat[0]
    return mat

def get_nimgs(cell, precision=None):
    r'''Choose number of basis function images in lattice sums
    to include for given precision in overlap, using

    precision ~ \int r^l e^{-\alpha r^2} (r-rcut)^l e^{-\alpha (r-rcut)^2}
    ~ (rcut^2/(2\alpha))^l e^{\alpha/2 rcut^2}

    where \alpha is the smallest exponent in the basis. Note
    that assumes an isolated exponent in the middle of the box, so
    it adds one additional lattice vector to be safe.
    '''
    if precision is None:
        precision = cell.precision

    # nimgs determines the supercell size
    rcut = estimate_rcut(cell, precision)
    nimgs = cell.get_bounding_sphere(rcut)
    return nimgs

def _estimate_rcut(alpha, l, c, precision=INTEGRAL_PRECISION):
    '''rcut based on the overlap integrals. This estimation is too conservative
    in many cases. A possible replacement can be the value of the basis
    function at rcut ~ c*r^(l+2)*exp(-alpha*r^2) < precision'''
    theta = alpha * .5
    a1 = (alpha * 2)**-.5
    norm_ang = (2*l+1)/(4*np.pi)
    fac = 2*np.pi*c**2*norm_ang / theta / precision
    r0 = 20
    # The estimation is enough for overlap. Errors are slightly larger for
    # kinetic operator. The error in kinetic integrals may be dominant.
    # For kinetic operator, basis becomes 2*a*r*|orig-basis>.
    # A penalty term 4*a^2*r^2 is included in the estimation.
    fac *= 4*alpha**2
    r0 = (np.log(fac * r0 * (r0*.5+a1)**(2*l+2) + 1.) / theta)**.5
    r0 = (np.log(fac * r0 * (r0*.5+a1)**(2*l+2) + 1.) / theta)**.5
    return r0

def bas_rcut(cell, bas_id, precision=None):
    r'''Estimate the largest distance between the function and its image to
    reach the precision in overlap

    precision ~ \int g(r-0) g(r-Rcut)
    '''
    if precision is None:
        precision = cell.precision
    l = cell.bas_angular(bas_id)
    es = cell.bas_exp(bas_id)
    cs = abs(cell._libcint_ctr_coeff(bas_id)).max(axis=1)
    rcut = _estimate_rcut(es, l, cs, precision)
    return rcut.max()

def estimate_rcut(cell, precision=None):
    '''Lattice-sum cutoff for the entire system'''
    if cell.nbas == 0:
        return 0.01
    if precision is None:
        precision = cell.precision
    if cell.use_loose_rcut:
        return cell.rcut_by_shells(precision).max()

    exps, cs = _extract_pgto_params(cell, 'min')
    ls = cell._bas[:,mole.ANG_OF]
    rcut = _estimate_rcut(exps, ls, cs, precision)
    return rcut.max()

def _estimate_ke_cutoff(alpha, l, c, precision=INTEGRAL_PRECISION, omega=0):
    '''Energy cutoff estimation for nuclear attraction integrals'''
    norm_ang = (2*l+1)/(4*np.pi)
    fac = 32*np.pi**2*(2*np.pi)**1.5 * c**2*norm_ang / (2*alpha)**(2*l+.5) / precision
    Ecut = 20.
    if omega <= 0:
        Ecut = np.log(fac * (Ecut*2)**(l-.5) + 1.) * 4*alpha
        Ecut = np.log(fac * (Ecut*2)**(l-.5) + 1.) * 4*alpha
    else:
        theta = 1./(1./(4*alpha) + 1./(2*omega**2))
        Ecut = np.log(fac * (Ecut*2)**(l-.5) + 1.) * theta
        Ecut = np.log(fac * (Ecut*2)**(l-.5) + 1.) * theta
    return Ecut

def estimate_ke_cutoff(cell, precision=None):
    '''Energy cutoff estimation for nuclear attraction integrals'''
    if cell.nbas == 0:
        return 0.
    if precision is None:
        precision = cell.precision
    #precision /= cell.atom_charges().sum()
    exps, cs = _extract_pgto_params(cell, 'max')
    ls = cell._bas[:,mole.ANG_OF]
    Ecut = _estimate_ke_cutoff(exps, ls, cs, precision, cell.omega)
    return Ecut.max()

def _extract_pgto_params(cell, op='min'):
    '''A helper function for estimate_xxx function'''
    es = []
    cs = []
    if op == 'min':
        for i in range(cell.nbas):
            e = cell.bas_exp(i)
            c = cell._libcint_ctr_coeff(i)
            idx = e.argmin()
            es.append(e[idx])
            cs.append(abs(c[idx]).max())
    else:
        for i in range(cell.nbas):
            e = cell.bas_exp(i)
            c = cell._libcint_ctr_coeff(i)
            idx = e.argmax()
            es.append(e[idx])
            cs.append(abs(c[idx]).max())
    return np.array(es), np.array(cs)

def error_for_ke_cutoff(cell, ke_cutoff, omega=None):
    '''Error estimation based on nuclear attraction integrals'''
    if omega is None:
        omega = cell.omega
    exps, cs = _extract_pgto_params(cell, 'max')
    ls = cell._bas[:,mole.ANG_OF]
    norm_ang = (2*ls+1)/(4*np.pi)
    fac = 32*np.pi**2*(2*np.pi)**1.5 * cs**2*norm_ang / (2*exps)**(2*ls+.5)
    if omega <= 0:
        err = fac * (2*ke_cutoff)**(ls-.5) * np.exp(-ke_cutoff/(4*exps))
    else:
        theta = 1./(1./(4*exps) + 1./(2*omega**2))
        err = fac * (2*ke_cutoff)**(ls-.5) * np.exp(-ke_cutoff/theta)
    return err.max()

def get_bounding_sphere(cell, rcut):
    '''Finds all the lattice points within a sphere of radius rcut.

    Defines a parallelepiped given by -N_x <= n_x <= N_x, with x in [1,3]
    See Martin p. 85

    Args:
        rcut : number
            real space cut-off for interaction

    Returns:
        cut : ndarray of 3 ints defining N_x
    '''
    #Gmat = cell.reciprocal_vectors(norm_to=1)
    #n1 = np.ceil(lib.norm(Gmat[0,:])*rcut)
    #n2 = np.ceil(lib.norm(Gmat[1,:])*rcut)
    #n3 = np.ceil(lib.norm(Gmat[2,:])*rcut)
    #cut = np.array([n1, n2, n3]).astype(int)
    b = cell.reciprocal_vectors(norm_to=1)
    heights_inv = lib.norm(b, axis=1)
    nimgs = np.ceil(rcut*heights_inv).astype(int)

    for i in range(cell.dimension, 3):
        nimgs[i] = 0
    return nimgs

def get_Gv(cell, mesh=None, **kwargs):
    '''Calculate three-dimensional G-vectors for the cell; see MH (3.8).

    Indices along each direction go as [0...N-1, -N...-1] to follow FFT convention.

    Args:
        cell : instance of :class:`Cell`

    Returns:
        Gv : (ngrids, 3) ndarray of floats
            The array of G-vectors.
    '''
    return get_Gv_weights(cell, mesh, **kwargs)[0]

def get_Gv_weights(cell, mesh=None, **kwargs):
    '''Calculate G-vectors and weights.

    Returns:
        Gv : (ngris, 3) ndarray of floats
            The array of G-vectors.
    '''
    if mesh is None:
        mesh = cell.mesh
    if 'gs' in kwargs:
        warnings.warn('cell.gs is deprecated.  It is replaced by cell.mesh,'
                      'the number of PWs (=2*gs+1) along each direction.')
        mesh = [2*n+1 for n in kwargs['gs']]

    # Default, the 3D uniform grids
    rx = np.fft.fftfreq(mesh[0], 1./mesh[0])
    ry = np.fft.fftfreq(mesh[1], 1./mesh[1])
    rz = np.fft.fftfreq(mesh[2], 1./mesh[2])
    b = cell.reciprocal_vectors()
    weights = abs(np.linalg.det(b))

    if cell.dimension <= 2 and cell.low_dim_ft_type == 'inf_vacuum':
        if cell.dimension == 0:
            rx, wx = _non_uniform_Gv_base(mesh[0]//2)
            ry, wy = _non_uniform_Gv_base(mesh[1]//2)
            rz, wz = _non_uniform_Gv_base(mesh[2]//2)
            rx /= np.linalg.norm(b[0])
            ry /= np.linalg.norm(b[1])
            rz /= np.linalg.norm(b[2])
            weights = np.einsum('i,j,k->ijk', wx, wy, wz).reshape(-1)
        elif cell.dimension == 1:
            wx = np.repeat(np.linalg.norm(b[0]), mesh[0])
            ry, wy = _non_uniform_Gv_base(mesh[1]//2)
            rz, wz = _non_uniform_Gv_base(mesh[2]//2)
            ry /= np.linalg.norm(b[1])
            rz /= np.linalg.norm(b[2])
            weights = np.einsum('i,j,k->ijk', wx, wy, wz).reshape(-1)
        elif cell.dimension == 2:
            area = np.linalg.norm(np.cross(b[0], b[1]))
            wxy = np.repeat(area, mesh[0]*mesh[1])
            rz, wz = _non_uniform_Gv_base(mesh[2]//2)
            rz /= np.linalg.norm(b[2])
            weights = np.einsum('i,k->ik', wxy, wz).reshape(-1)

    Gvbase = (rx, ry, rz)

    #:Gv = np.dot(lib.cartesian_prod(Gvbase), b)
    # NOTE mesh can be different from the input mesh
    mesh = np.asarray((len(rx),len(ry),len(rz)), dtype=np.int32)
    Gv = np.empty((*mesh,3), order='C', dtype=float)
    b = np.asarray(b, order='C')
    rx = np.asarray(rx, order='C')
    ry = np.asarray(ry, order='C')
    rz = np.asarray(rz, order='C')
    libpbc.get_Gv(
        Gv.ctypes.data_as(ctypes.c_void_p),
        rx.ctypes.data_as(ctypes.c_void_p),
        ry.ctypes.data_as(ctypes.c_void_p),
        rz.ctypes.data_as(ctypes.c_void_p),
        mesh.ctypes.data_as(ctypes.c_void_p),
        b.ctypes.data_as(ctypes.c_void_p),
    )
    Gv = Gv.reshape(-1, 3)

    # 1/cell.vol == det(b)/(2pi)^3
    weights *= 1/(2*np.pi)**3
    return Gv, Gvbase, weights

def _non_uniform_Gv_base(n):
    #rs, ws = radi.delley(n)
    #rs, ws = radi.treutler_ahlrichs(n)
    #rs, ws = radi.mura_knowles(n)
    rs, ws = radi.gauss_chebyshev(n)
    #return np.hstack((0,rs,-rs[::-1])), np.hstack((0,ws,ws[::-1]))
    return np.hstack((rs,-rs[::-1])), np.hstack((ws,ws[::-1]))

def get_SI(cell, Gv=None, mesh=None, atmlst=None):
    '''Calculate the structure factor (0D, 1D, 2D, 3D) for all atoms; see MH (3.34).

    Args:
        cell : instance of :class:`Cell`

        Gv : (N,3) array
            G vectors

        atmlst : list of ints, optional
            Indices of atoms for which the structure factors are computed.

    Returns:
        SI : (natm, ngrids) ndarray, dtype=np.complex128
            The structure factor for each atom at each G-vector.
    '''
    coords = cell.atom_coords()
    if atmlst is not None:
        coords = coords[np.asarray(atmlst)]
    if Gv is None:
        if mesh is None:
            mesh = cell.mesh
        basex, basey, basez = cell.get_Gv_weights(mesh)[1]
        b = cell.reciprocal_vectors()
        rb = np.dot(coords, b.T)
        SIx = np.exp(-1j*np.einsum('z,g->zg', rb[:,0], basex))
        SIy = np.exp(-1j*np.einsum('z,g->zg', rb[:,1], basey))
        SIz = np.exp(-1j*np.einsum('z,g->zg', rb[:,2], basez))
        SI = SIx[:,:,None,None] * SIy[:,None,:,None] * SIz[:,None,None,:]
        natm = coords.shape[0]
        SI = SI.reshape(natm, -1)
    else:
        SI = np.exp(-1j*np.dot(coords, Gv.T))
    return SI

def get_ewald_params(cell, precision=None, mesh=None):
    r'''Choose a reasonable value of Ewald 'eta' and 'cut' parameters.
    eta^2 is the exponent coefficient of the model Gaussian charge for nucleus
    at R:  \frac{eta^3}{pi^1.5} e^{-eta^2 (r-R)^2}

    Choice is based on largest G vector and desired relative precision.

    The relative error in the G-space sum is given by

        precision ~ 4\pi Gmax^2 e^{(-Gmax^2)/(4 \eta^2)}

    which determines eta. Then, real-space cutoff is determined by (exp.
    factors only)

        precision ~ erfc(eta*rcut) / rcut ~ e^{(-eta**2 rcut*2)}

    Returns:
        ew_eta, ew_cut : float
            The Ewald 'eta' and 'cut' parameters.
    '''
    if cell.natm == 0:
        return 0, 0

    if precision is None:
        precision = cell.precision

    if (cell.dimension == 3 or
        (cell.dimension == 0 and cell.low_dim_ft_type != 'inf_vacuum')):
        ew_eta = 1./cell.vol**(1./6)
        ew_cut = _estimate_rcut(ew_eta**2, 0, 1., precision)
    elif cell.dimension <= 2 and cell.low_dim_ft_type == 'inf_vacuum':
        # Non-uniform PW grids are used for low-dimensional ewald summation.  The cutoff
        # estimation for long range part based on exp(G^2/(4*eta^2)) does not work for
        # non-uniform grids.  Smooth model density is preferred.
        ew_cut = cell.rcut
        ew_eta = np.sqrt(max(np.log(4*np.pi*ew_cut**2/precision)/ew_cut**2, .1))
    elif cell.dimension == 2:
        a = cell.lattice_vectors()
        ew_cut = a[2,2] / 2
        # ewovrl ~ erfc(eta*rcut) / rcut ~ e^{(-eta**2 rcut*2)} < precision
        log_precision = np.log(precision / (cell.atom_charges().sum()*16*np.pi**2))
        ew_eta = (-log_precision)**.5 / ew_cut
    else:
        raise RuntimeError(f'dimension={cell.dimension} not supported')
    return ew_eta, ew_cut

def ewald(cell, ew_eta=None, ew_cut=None):
    '''Perform real (R) and reciprocal (G) space Ewald sum for the energy.

    Formulation of Martin, App. F2.

    Returns:
        float
            The Ewald energy consisting of overlap, self, and G-space sum.

    See Also:
        pyscf.pbc.gto.get_ewald_params
    '''
    # If lattice parameter is not set, the cell object is treated as a mole
    # object. The nuclear repulsion energy is computed.
    if cell.a is None:
        return mole.energy_nuc(cell)

    if cell.natm == 0:
        return 0

    if cell.dimension == 3 and cell.use_particle_mesh_ewald:
        from pyscf.pbc.gto import ewald_methods
        return ewald_methods.particle_mesh_ewald(cell, ew_eta, ew_cut)

    if cell.dimension == 0 and cell.low_dim_ft_type != 'inf_vacuum':
        return mole.classical_coulomb_energy(cell)

    chargs = cell.atom_charges()

    if ew_eta is None or ew_cut is None:
        ew_eta, ew_cut = cell.get_ewald_params()
    log_precision = np.log(cell.precision / (chargs.sum()*16*np.pi**2))
    ke_cutoff = -2*ew_eta**2*log_precision
    mesh = cell.cutoff_to_mesh(ke_cutoff)
    logger.debug1(cell, 'mesh for ewald %s', mesh)

    coords = cell.atom_coords()
    Lall = cell.get_lattice_Ls(rcut=ew_cut)

    rLij = coords[:,None,:] - coords[None,:,:] + Lall[:,None,None,:]
    r = np.sqrt(np.einsum('Lijx,Lijx->Lij', rLij, rLij))
    rLij = None
    r[r<1e-16] = 1e200
    ewovrl = .5 * np.einsum('i,j,Lij->', chargs, chargs, erfc(ew_eta * r) / r)

    # last line of Eq. (F.5) in Martin
    ewself  = -.5 * np.dot(chargs,chargs) * 2 * ew_eta / np.sqrt(np.pi)
    if cell.dimension == 3:
        ewself += -.5 * np.sum(chargs)**2 * np.pi/(ew_eta**2 * cell.vol)

    # g-space sum (using g grid) (Eq. (F.6) in Martin, but note errors as below)
    # Eq. (F.6) in Martin is off by a factor of 2, the
    # exponent is wrong (8->4) and the square is in the wrong place
    #
    # Formula should be
    #   1/2 * 4\pi / Omega \sum_I \sum_{G\neq 0} |ZS_I(G)|^2 \exp[-|G|^2/4\eta^2]
    # where
    #   ZS_I(G) = \sum_a Z_a exp (i G.R_a)

    if cell.dimension == 3 or cell.low_dim_ft_type == 'inf_vacuum':
        Gv, Gvbase, weights = cell.get_Gv_weights(mesh)
        absG2 = np.einsum('gi,gi->g', Gv, Gv)
        absG2[absG2==0] = 1e200
        coulG = 4*np.pi / absG2
        coulG *= weights

        #:ZSI = np.einsum('i,ij->j', chargs, cell.get_SI(Gv))
        basex, basey, basez = Gvbase
        b = cell.reciprocal_vectors()
        rb = np.dot(coords, b.T)
        SIx = np.exp(-1j*np.einsum('z,g->zg', rb[:,0], basex))
        SIy = np.exp(-1j*np.einsum('z,g->zg', rb[:,1], basey))
        SIz = np.exp(-1j*np.einsum('z,g->zg', rb[:,2], basez))
        ZSI = np.einsum('i,ix,iy,iz->xyz', chargs, SIx, SIy, SIz).ravel()
        ZexpG2 = ZSI * np.exp(-absG2/(4*ew_eta**2))
        ewg = .5 * np.einsum('i,i,i', ZSI.conj(), ZexpG2, coulG).real

    elif cell.dimension == 2:  # Truncated Coulomb
        # The following 2D ewald summation is taken from:
        # R. Sundararaman and T. Arias PRB 87, 2013
        def fn(eta,Gnorm,z):
            Gnorm_z = Gnorm*z
            large_idx = Gnorm_z > 20.0
            ret = np.zeros_like(Gnorm_z)
            x = Gnorm/2./eta + eta*z
            with np.errstate(over='ignore'):
                erfcx = erfc(x)
                ret[~large_idx] = np.exp(Gnorm_z[~large_idx]) * erfcx[~large_idx]
                ret[ large_idx] = np.exp((Gnorm*z-x**2)[large_idx]) * erfcx[large_idx]
            return ret
        def gn(eta,Gnorm,z):
            return np.pi/Gnorm*(fn(eta,Gnorm,z) + fn(eta,Gnorm,-z))
        def gn0(eta,z):
            return -2*np.pi*(z*erf(eta*z) + np.exp(-(eta*z)**2)/eta/np.sqrt(np.pi))
        b = cell.reciprocal_vectors()
        inv_area = np.linalg.norm(np.cross(b[0], b[1]))/(2*np.pi)**2
        # Perform the reciprocal space summation over  all reciprocal vectors
        # within the x,y plane.
        Gv, Gvbase, weights = cell.get_Gv_weights(mesh)
        absG2 = np.einsum('gi,gi->g', Gv, Gv)
        planarG2_idx = np.logical_and(Gv[:,2] == 0, absG2 > 0.0)
        Gv = Gv[planarG2_idx]
        absG2 = absG2[planarG2_idx]
        absG = absG2**(0.5)
        # Performing the G != 0 summation.
        rij = coords[:,None,:] - coords[None,:,:]
        Gdotr = np.einsum('ijx,gx->ijg', rij, Gv)
        ewg = np.einsum('i,j,ijg,ijg->', chargs, chargs, np.cos(Gdotr),
                        gn(ew_eta,absG,rij[:,:,2:3]))
        # Performing the G == 0 summation.
        ewg += np.einsum('i,j,ij->', chargs, chargs, gn0(ew_eta,rij[:,:,2]))
        ewg *= inv_area*0.5

    elif cell.dimension == 0:
        # The truncted Coulomb kernel 4*pi/G^2(1-cos(G*Rc)) should work for the
        # ewg term. However, large errors may be introduced in ewovrl or ewself.
        # For dimesion=0 with truncated Coulomb, the nuclear energy can be
        # computed directly using the mole.classical_coulomb_energy function.
        raise RuntimeError('Ewald with truncated Coulomb')

    else:
        logger.warn(cell, 'No method for PBC dimension %s, dim-type %s.'
                    '  cell.low_dim_ft_type="inf_vacuum"  should be set.',
                    cell.dimension, cell.low_dim_ft_type)
        raise NotImplementedError

    logger.debug(cell, 'Ewald components = %.15g, %.15g, %.15g', ewovrl, ewself, ewg)
    return ewovrl + ewself + ewg

energy_nuc = ewald

def make_kpts(cell, nks, wrap_around=WRAP_AROUND, with_gamma_point=WITH_GAMMA,
              scaled_center=None,
              space_group_symmetry=False, time_reversal_symmetry=False,
              **kwargs):
    '''Given number of kpoints along x,y,z , generate kpoints

    Args:
        nks : (3,) ndarray

    Kwargs:
        wrap_around : bool
            To ensure all kpts are in first Brillouin zone.
        with_gamma_point : bool
            Whether to shift Monkhorst-pack grid to include gamma-point.
        scaled_center : (3,) array
            Shift all points in the Monkhorst-pack grid to be centered on
            scaled_center, given as the zeroth index of the returned kpts.
            Scaled meaning that the k-points are scaled to a grid from
            [-1,1] x [-1,1] x [-1,1]
        space_group_symmetry : bool
            Whether to consider space group symmetry
        time_reversal_symmetry : bool
            Whether to consider time reversal symmetry

    Returns:
        kpts in absolute value (unit 1/Bohr).  Gamma point is placed at the
        first place in the k-points list;
        instance of :class:`KPoints` if k-point symmetry is considered

    Examples:

    >>> cell.make_kpts((4,4,4))
    '''
    ks_each_axis = []
    for n in nks:
        if with_gamma_point or scaled_center is not None:
            ks = np.arange(n, dtype=float) / n
        else:
            ks = (np.arange(n)+.5)/n-.5
        if wrap_around:
            ks[ks>=.5] -= 1
        ks_each_axis.append(ks)
    if scaled_center is None:
        scaled_center = [0.0,0.0,0.0]
    scaled_kpts = lib.cartesian_prod(ks_each_axis)
    scaled_kpts += np.array(scaled_center)
    kpts = cell.get_abs_kpts(scaled_kpts)
    if space_group_symmetry or time_reversal_symmetry:
        from pyscf.pbc.lib import kpts as libkpts
        if space_group_symmetry and not cell.space_group_symmetry:
            raise RuntimeError('Using k-point symmetry now requires cell '
                               'to be built with space group symmetry info:\n'
                               'cell.space_group_symmetry = True\n'
                               'cell.symmorphic = False\n'
                               'cell.build()')
        kpts = libkpts.make_kpts(cell, kpts, space_group_symmetry,
                                 time_reversal_symmetry)
    return kpts

def get_uniform_grids(cell, mesh=None, wrap_around=True):
    '''Generate a uniform real-space grid consistent w/ samp thm; see MH (3.19).

    Args:
        cell : instance of :class:`Cell`

    Returns:
        coords : (ngx*ngy*ngz, 3) ndarray
            The real-space grid point coordinates.

    '''
    if mesh is None: mesh = cell.mesh

    if wrap_around:
        # wrap the coordinates around the origin. If coordinates are generated
        # inside the primitive cell without wrap-around, an extra layer would be
        # needed in function get_lattice_Ls for 2D calculations.
        qv = lib.cartesian_prod([np.fft.fftfreq(x) for x in mesh])
        coords = np.dot(qv, cell.lattice_vectors())
    else:
        mesh = np.asarray(mesh, dtype=np.double)
        qv = lib.cartesian_prod([np.arange(x) for x in mesh])
        a_frac = np.einsum('i,ij->ij', 1./mesh, cell.lattice_vectors())
        coords = np.dot(qv, a_frac)
    return coords
gen_uniform_grids = get_uniform_grids

def _split_basis(cell, delimiter=EXP_DELIMITER):
    '''
    Split the contracted basis to small segmant.  The new basis has more
    shells.  Each shell has less primitive basis and thus is more local.
    '''
    _bas = []
    _env = cell._env.copy()
    contr_coeff = []
    for ib in range(cell.nbas):
        pexp = cell._bas[ib,mole.PTR_EXP]
        pcoeff1 = cell._bas[ib,mole.PTR_COEFF]
        nc = cell.bas_nctr(ib)
        es = cell.bas_exp(ib)
        cs = cell._libcint_ctr_coeff(ib)
        l = cell.bas_angular(ib)
        if cell.cart:
            degen = (l + 1) * (l + 2) // 2
        else:
            degen = l * 2 + 1

        mask = np.ones(es.size, dtype=bool)
        count = 0
        for thr in delimiter:
            idx = np.where(mask & (es >= thr))[0]
            np1 = len(idx)
            if np1 > 0:
                pcoeff0, pcoeff1 = pcoeff1, pcoeff1 + np1 * nc
                cs1 = cs[idx]
                _env[pcoeff0:pcoeff1] = cs1.T.ravel()
                btemp = cell._bas[ib].copy()
                btemp[mole.NPRIM_OF] = np1
                btemp[mole.PTR_COEFF] = pcoeff0
                btemp[mole.PTR_EXP] = pexp
                _bas.append(btemp)
                mask[idx] = False
                pexp += np1
                count += 1
        contr_coeff.append(np.vstack([np.eye(degen*nc)] * count))

    pcell = cell.copy(deep=False)
    pcell._bas = np.asarray(np.vstack(_bas), dtype=np.int32)
    pcell._env = _env
    return pcell, scipy.linalg.block_diag(*contr_coeff)

def tot_electrons(cell, nkpts=1):
    '''Total number of electrons
    '''
    if cell._nelectron is None:
        nelectron = cell.atom_charges().sum() * nkpts - cell.charge
    else: # Custom cell.nelectron stands for num. electrons per cell
        nelectron = cell._nelectron * nkpts
    # Round off to the nearest integer
    nelectron = int(nelectron+0.5)
    return nelectron

def _mesh_inf_vaccum(cell):
    #prec ~ exp(-0.436392335*mesh -2.99944305)*nelec
    meshz = (np.log(cell.nelectron/cell.precision)-2.99944305)/0.436392335
    # meshz has to be even number due to the symmetry on z+ and z-
    return int(meshz*.5 + .999) * 2

def pgf_rcut(l, alpha, coeff, precision=INTEGRAL_PRECISION,
             rcut=0, max_cycle=RCUT_MAX_CYCLE, eps=RCUT_EPS):
    '''Estimate the cutoff radii of primitive Gaussian functions
    based on their values in real space:
    `c*rcut^(l+2)*exp(-alpha*rcut^2) ~ precision`.
    '''
    c = np.log(coeff / precision)

    rmin = np.sqrt(.5 * (l+2) / alpha) * 2
    eps = np.minimum(rmin/10, eps)
    rcut = np.maximum(rcut, rmin+eps)
    for i in range(max_cycle):
        rcut_last = rcut
        rcut = np.sqrt(((l+2) * np.log(rcut) + c) / alpha)
        if np.all(abs(rcut - rcut_last) < eps):
            return rcut
    warnings.warn(f'cell.pgf_rcut failed to converge in {max_cycle} cycles.')
    return rcut

def rcut_by_shells(cell, precision=None, rcut=0,
                   return_pgf_radius=False):
    '''Compute shell and primitive gaussian function radii.
    '''
    # TODO the internal implementation loops over all shells,
    # which can be optimized to loop over atom types.
    if precision is None:
        precision = cell.precision

    bas = np.asarray(cell._bas, order='C')
    env = np.asarray(cell._env, order='C')
    nbas = len(bas)
    shell_radius = np.empty((nbas,), order='C', dtype=float)
    if return_pgf_radius:
        nprim = bas[:,mole.NPRIM_OF].max()
        # be careful that the unused memory blocks are not initialized
        pgf_radius = np.empty((nbas,nprim), order='C', dtype=np.double)
        ptr_pgf_radius = lib.ndarray_pointer_2d(pgf_radius).ctypes
    else:
        ptr_pgf_radius = lib.c_null_ptr()
    libpbc.rcut_by_shells(
        shell_radius.ctypes.data_as(ctypes.c_void_p),
        ptr_pgf_radius,
        bas.ctypes.data_as(ctypes.c_void_p),
        env.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(nbas),
        ctypes.c_double(rcut),
        ctypes.c_double(precision),
    )
    if return_pgf_radius:
        return shell_radius, pgf_radius
    return shell_radius

def tostring(cell, format='poscar'):
    '''Convert cell geometry to a string of the required format.

    Supported output formats:
        | poscar: VASP POSCAR
        | xyz: Extended XYZ with Lattice information
    '''
    format = format.lower()
    output = []
    atmfmt = '%17.8f %17.8f %17.8f'
    if format == 'poscar' or format == 'vasp' or format == 'xyz':
        lattice_vectors = cell.lattice_vectors() * param.BOHR
        coords = cell.atom_coords() * param.BOHR
        if format == 'poscar' or format == 'vasp':
            output.append('Written by PySCF, units are A')
            output.append('1.0')
            for lattice_vector in lattice_vectors:
                ax, ay, az = lattice_vector
                output.append(atmfmt % (ax, ay, az))
            unique_atoms = dict()
            for atom in cell.elements:
                if atom not in unique_atoms:
                    unique_atoms[atom] = 1
                else:
                    unique_atoms[atom] += 1
            output.append(' '.join(unique_atoms))
            output.append(' '.join(str(count) for count in unique_atoms.values()))
            output.append('Cartesian')
            for atom_type in unique_atoms:
                for atom, coord in zip(cell.elements, coords):
                    if atom == atom_type:
                        x, y, z = coord
                        output.append(atmfmt % (x, y, z))
            return '\n'.join(output)
        elif format == 'xyz':
            output.append('%d' % cell.natm)
            output.append('Lattice="'+' '.join(f'{ax:17.8f}' for ax in lattice_vectors.ravel())
                +'" Properties=species:S:1:pos:R:3')
            for i in range(cell.natm):
                symb = cell.atom_pure_symbol(i)
                x, y, z = coords[i]
                output.append(('%-4s ' + atmfmt) %
                              (symb, x, y, z))
            return '\n'.join(output)
    else:
        raise NotImplementedError(f'format={format}')

def tofile(cell, filename, format=None):
    if format is None:  # Guess format based on filename
        if filename.lower() == 'poscar':
            format = 'poscar'
        else:
            format = os.path.splitext(filename)[1][1:]
    string = tostring(cell, format)
    with open(filename,  'w', encoding='utf-8') as f:
        f.write(string)
        f.write('\n')
    return string

def fromfile(filename, format=None):
    '''Read cell geometry from a file
    (in testing)

    Supported formats:
        | poscar: VASP POSCAR file format
        | xyz: Extended XYZ with Lattice information
    '''
    if format is None:  # Guess format based on filename
        if filename.lower() == 'poscar':
            format = 'poscar'
        else:
            format = os.path.splitext(filename)[1][1:].lower()
        if format not in ('poscar', 'vasp', 'xyz'):
            format = 'raw'
    with open(filename, 'r') as f:
        return fromstring(f.read(), format)

def fromstring(string, format='poscar'):
    '''Convert the string of the specified format to internal format
    (in testing)

    Supported formats:
        | poscar: VASP POSCAR file format
        | xyz: Extended XYZ with Lattice information

    Returns:
        a: Lattice vectors
        atom: Atomic elements and xyz coordinates
    '''
    format = format.lower()
    if format == 'poscar' or format == 'vasp':
        lines = string.splitlines()
        scale = float(lines[1])
        a = lines[2:5]
        lattice_vectors = np.array([np.fromstring(ax, sep=' ') for ax in a])
        lattice_vectors *= scale
        a = []
        for i in range(3):
            a.append(' '.join(str(ax) for ax in lattice_vectors[i]))
        atom_position_type = lines[7].strip()
        unique_atoms = dict()
        natm = 0
        for atom, count in zip(lines[5].split(), lines[6].split()):
            unique_atoms[atom] = int(count)
            natm += int(count)
        atoms = []
        start = 8
        for atom_type in unique_atoms:
            end = start + unique_atoms[atom_type]
            for line in lines[start:end]:
                coords = np.fromstring(line, sep=' ')
                if atom_position_type.lower() == 'cartesian':
                    x, y, z = coords * scale
                elif atom_position_type.lower() == 'direct':
                    x, y, z = np.dot(coords, lattice_vectors)
                else:
                    raise RuntimeError('Error reading VASP geometry due to '
                        f'atom position type "{atom_position_type}". Atom '
                        'positions must be Direct or Cartesian.')
                atoms.append('%s %17.8f %17.8f %17.8f'
                    % (atom_type, x, y, z))
            start = end
        return '\n'.join(a), '\n'.join(atoms)
    elif format == 'xyz':
        lines = string.splitlines()
        natm = int(lines[0])
        lattice_vectors = lines[1].split('Lattice=')[1].split('"')[1].split()
        a = []
        for i in range(3):
            a.append(" ".join(lattice_vectors[3*i:3*i+3]))
        return '\n'.join(a), '\n'.join(lines[2:natm+2])
    elif format == 'raw':
        lines = string.splitlines()
        return '\n'.join(lines[:3]), '\n'.join(lines[4:])
    else:
        raise NotImplementedError


class Cell(mole.MoleBase):
    '''A Cell object holds the basic information of a crystal.

    Attributes:
        a : (3,3) ndarray
            Lattice primitive vectors. Each row represents a lattice vector
            Reciprocal lattice vectors are given by  b1,b2,b3 = 2 pi inv(a).T
        mesh : (3,) list of ints
            The number G-vectors along each direction.
            The default value is estimated based on :attr:`precision`
        pseudo : dict or str
            To define pseudopotential.
        precision : float
            To control Ewald sums and lattice sums accuracy
        rcut : float
            Cutoff radius (unit Bohr) in lattice summation. The default value
            is estimated based on the required :attr:`precision`.
        ke_cutoff : float
            If set, defines a spherical cutoff of planewaves, with .5 * G**2 < ke_cutoff
            The default value is estimated based on :attr:`precision`
        dimension : int
            Periodic dimensions. Default is 3
        low_dim_ft_type : str
            For semi-empirical periodic systems, whether to calculate
            integrals at the non-PBC dimension using the sampled mesh grids in
            infinity vacuum (inf_vacuum) or truncated Coulomb potential
            (analytic_2d_1). Unless explicitly specified, analytic_2d_1 is
            used for 2D system and inf_vacuum is assumed for 1D and 0D.
        use_loose_rcut : bool
            If set to True, a loose `rcut` determined by shell radius is used,
            which is usually accurate enough for pure DFT calculations;
            otherwise, a tight `rcut` determined by overlap integral is used.
            Default value is False. Has no effect if `rcut` is set manually.
        use_particle_mesh_ewald : bool
            If set to True, use particle-mesh Ewald to compute the nuclear repulsion.
            Default value is False, meaning to use classical Ewald summation.
        space_group_symmetry : bool
            Whether to consider space group symmetry. Default is False.
        symmorphic : bool
            Whether the lattice is symmorphic. If set to True, even if the
            lattice is non-symmorphic, only symmorphic space group symmetry
            will be considered. Default is False, meaning the space group is
            determined by the lattice symmetry to be symmorphic or non-symmorphic.
        lattice_symmetry : None or :class:`pbc.symm.Symmetry` instance
            The object containing the lattice symmetry information. Default is None.

    (See other attributes in :class:`Mole`)

    Examples:

    >>> mol = Mole(atom='H^2 0 0 0; H 0 0 1.1', basis='sto3g')
    >>> cl = Cell()
    >>> cl.build(a='3 0 0; 0 3 0; 0 0 3', atom='C 1 1 1', basis='sto3g')
    >>> print(cl.atom_symbol(0))
    C
    '''

    precision = getattr(__config__, 'pbc_gto_cell_Cell_precision', 1e-8)
    exp_to_discard = getattr(__config__, 'pbc_gto_cell_Cell_exp_to_discard', None)

    fractional = False
    dimension = 3
    # TODO: Simple hack for now; the implementation of ewald depends on the
    #       density-fitting class.  This determines how the ewald produces
    #       its energy.
    low_dim_ft_type = None
    use_loose_rcut = False
    use_particle_mesh_ewald = False
    space_group_symmetry = False
    symmorphic = False
    lattice_symmetry = None

    _keys = {
        'precision', 'exp_to_discard',
        'a', 'ke_cutoff', 'pseudo', 'fractional', 'dimension', 'low_dim_ft_type',
        'space_group_symmetry', 'symmorphic', 'lattice_symmetry', 'mesh', 'rcut',
        'use_loose_rcut', 'use_particle_mesh_ewald',
    }

    tostring = tostring
    tofile = tofile

    def __init__(self, **kwargs):
        mole.MoleBase.__init__(self)
        self.a = None # lattice vectors, (a1,a2,a3)
        # if set, defines a spherical cutoff
        # of fourier components, with .5 * G**2 < ke_cutoff
        self.ke_cutoff = None

##################################################
# These attributes are initialized by build function if not specified
        self.mesh = None
        self.rcut = None
        for key, val in kwargs.items():
            setattr(self, key, val)

    def fromstring(self, string, format='poscar'):
        '''Update the Cell object based on the input geometry string'''
        a, atom = fromstring(string, format)
        self.a = a
        self.set_geom_(atom, unit='Angstrom', inplace=True)
        return self

    def fromfile(self, filename, format=None):
        '''Update the Cell object based on the input geometry file'''
        a, atom = fromfile(filename, format)
        self.a = a
        self.set_geom_(atom, unit='Angstrom', inplace=True)
        return self

    @property
    def mesh(self):
        return self._mesh
    @mesh.setter
    def mesh(self, x):
        self._mesh = x
        self._mesh_from_build = False

    @property
    def rcut(self):
        return self._rcut
    @rcut.setter
    def rcut(self, x):
        self._rcut = x
        self._rcut_from_build = False

    @property
    def ew_eta(self):
        warnings.warn("cell.ew_eta is deprecated. Use function get_ewald_params instead.")
        return self.get_ewald_params()[0]

    @property
    def ew_cut(self):
        warnings.warn("cell.ew_cut is deprecated. Use function get_ewald_params instead.")
        return self.get_ewald_params()[1]

    @ew_eta.setter
    def ew_eta(self, val):
        warnings.warn("ew_eta is no longer stored in the cell object. Setting it has no effect")

    @ew_cut.setter
    def ew_cut(self, val):
        warnings.warn("ew_cut is no longer stored in the cell object. Setting it has no effect")

    @property
    def nelec(self):
        ne = self.nelectron
        nalpha = (ne + self.spin) // 2
        nbeta = nalpha - self.spin
        # nelec method defined in Mole class raises error when the attributes .spin
        # and .nelectron are inconsistent.  In PBC, when the system has even number of
        # k-points, it is valid that .spin is odd while .nelectron is even.
        if nalpha + nbeta != ne:
            warnings.warn('Electron number %d and spin %d are not consistent '
                          'in cell\n' % (ne, self.spin))
        return nalpha, nbeta

    def __getattr__(self, key):
        '''To support accessing methods (cell.HF, cell.KKS, cell.KUCCSD, ...)
        from Cell object.
        '''
        if key[0] == '_':  # Skip private attributes and Python builtins
            # https://bugs.python.org/issue45985
            # https://github.com/python/cpython/issues/103936
            # @property and __getattr__ conflicts. As a temporary fix, call
            # object.__getattribute__ method to re-raise AttributeError
            return object.__getattribute__(self, key)

        # Import all available modules. Some methods are registered to other
        # classes/modules when importing modules in __all__.
        from pyscf.pbc import __all__  # noqa
        from pyscf.pbc import scf, dft
        from pyscf.dft import XC

        for mod in (scf, dft):
            method = getattr(mod, key, None)
            if callable(method):
                return method(self)

        if key[0] == 'K':  # with k-point sampling
            if 'TD' in key[:4]:
                if key in ('KTDHF', 'KTDA'):
                    mf = scf.KHF(self)
                else:
                    mf = dft.KKS(self)
                    xc = key.split('TD', 1)[1]
                    if xc in XC:
                        mf.xc = xc
                        key = 'KTDDFT'
            elif 'CI' in key or 'CC' in key or 'MP' in key:
                mf = scf.KHF(self)
            else:
                return object.__getattribute__(self, key)
            # Remove prefix 'K' because methods are registered without the leading 'K'
            key = key[1:]
        else:
            if 'TD' in key[:3]:
                if key in ('TDHF', 'TDA'):
                    mf = scf.HF(self)
                else:
                    mf = dft.KS(self)
                    xc = key.split('TD', 1)[1]
                    if xc in XC:
                        mf.xc = xc
                        key = 'TDDFT'
            elif 'CI' in key or 'CC' in key or 'MP' in key:
                mf = scf.HF(self)
            else:
                return object.__getattribute__(self, key)

        method = getattr(mf, key)

        if self.nelectron != 0:
            mf.run()
        return method

    tot_electrons = tot_electrons

    def _build_symmetry(self, kpts=None, **kwargs):
        '''Construct symmetry adapted crystalline atomic orbitals
        '''
        from pyscf.pbc.lib.kpts import KPoints
        from pyscf.pbc.symm.basis import symm_adapted_basis
        if kpts is None:
            return mole.Mole._build_symmetry(self)
        elif isinstance(kpts, KPoints):
            self.symm_orb, self.irrep_id = symm_adapted_basis(self, kpts)
            return self
        else:
            raise RuntimeError('Symmetry information not found in kpts. '
                               'kpts needs to be initialized as a KPoints object.')

    def symmetrize_mesh(self, mesh=None):
        if mesh is None:
            mesh = self.mesh
        if not self.space_group_symmetry:
            return mesh

        _, mesh1 = self.lattice_symmetry.check_mesh_symmetry(mesh=mesh,
                                                             return_mesh=True)
        if np.prod(mesh1) != np.prod(mesh):
            logger.debug(self, 'mesh %s is symmetrized as %s', mesh, mesh1)
        m1size = np.prod(mesh1)
        msize = np.prod(mesh)
        if m1size > 8 * msize and m1size > 1000 + msize:
            wstr = ('''WARNING!
  Symmetrization significantly increased the mesh size,
  from {} to {}. This might indicate a nearly symmetric input
  structure, and it might cause memory issues. Consider symmetrizing your
  structure, increasing the symmetry tolerance `pbc_symm_space_group_symprec`,
  or turning off symmetry.\n\n''')
            wstr = wstr.format(mesh, mesh1)
            sys.stderr.write(wstr)
        return mesh1

    def build_lattice_symmetry(self, check_mesh_symmetry=True):
        '''Build cell.lattice_symmetry object.

        Kwargs:
            check_mesh_symmetry : bool
                For nonsymmorphic symmetry groups, `cell.mesh` may have
                lower symmetry than the lattice. In this case, if
                `check_mesh_symmetry` is `True`, the lower symmetry group will
                be used. Otherwise, if `check_mesh_symmetry` is `False`,
                the mesh grid will be modified to satisfy the higher symmetry.
                Default value is `True`.

        Note:
            This function modifies the attributes of `cell`.
        '''
        from pyscf.pbc.symm import Symmetry
        self.lattice_symmetry = Symmetry(self).build(
                                    space_group_symmetry=True,
                                    symmorphic=self.symmorphic,
                                    check_mesh_symmetry=check_mesh_symmetry)
        if not check_mesh_symmetry:
            _mesh_from_build = self._mesh_from_build
            self.mesh = self.symmetrize_mesh()
            self._mesh_from_build = _mesh_from_build
        return self

    @lib.with_doc(mole.format_atom.__doc__)
    def format_atom(self, atoms, origin=0, axes=None,
                    unit=getattr(__config__, 'UNIT', 'Ang')):
        if not self.fractional:
            return mole.format_atom(atoms, origin, axes, unit)
        _atoms = mole.format_atom(atoms, origin, axes, unit=1.)
        _a = self.lattice_vectors()
        c = np.array([a[1] for a in _atoms]).dot(_a)
        return [(a[0], r) for a, r in zip(_atoms, c.tolist())]

#Note: Exculde dump_input, parse_arg, basis from kwargs to avoid parsing twice
    def build(self, dump_input=True, parse_arg=mole.ARGPARSE,
              a=None, mesh=None, ke_cutoff=None, precision=None, nimgs=None,
              h=None, dimension=None, rcut= None, low_dim_ft_type=None,
              space_group_symmetry=None, symmorphic=None,
              use_loose_rcut=None, use_particle_mesh_ewald=None,
              fractional=None, *args, **kwargs):
        '''Setup Mole molecule and Cell and initialize some control parameters.
        Whenever you change the value of the attributes of :class:`Cell`,
        you need call this function to refresh the internal data of Cell.

        Kwargs:
            a : (3,3) ndarray
                The real-space cell lattice vectors. Each row represents
                a lattice vector.
            fractional : bool
                Whether the atom postions are specified in fractional coordinates.
                The default value is False, which means the coordinates are
                interpreted as Cartesian coordinate.
            mesh : (3,) ndarray of ints
                The number of *positive* G-vectors along each direction.
            ke_cutoff : float
                If set, defines a spherical cutoff of planewaves, with .5 * G**2 < ke_cutoff
                The default value is estimated based on :attr:`precision`
            precision : float
                To control Ewald sums and lattice sums accuracy
            nimgs : (3,) ndarray of ints
                Number of repeated images in lattice summation to produce
                periodicity. This value can be estimated based on the required
                precision.  It's recommended NOT making changes to this value.
            rcut : float
                Cutoff radius (unit Bohr) in lattice summation to produce
                periodicity. The value can be estimated based on the required
                precision.  It's recommended NOT making changes to this value.
            h : (3,3) ndarray
                a.T. Deprecated
            dimension : int
                Default is 3
            low_dim_ft_type : str
                For semi-empirical periodic systems, whether to calculate
                integrals at the non-PBC dimension using the sampled mesh grids in
                infinity vacuum (inf_vacuum) or truncated Coulomb potential
                (analytic_2d_1). Unless explicitly specified, analytic_2d_1 is
                used for 2D system and inf_vacuum is assumed for 1D and 0D.
            space_group_symmetry : bool
                Whether to consider space group symmetry. Default is False.
            symmorphic : bool
                Whether the lattice is symmorphic. If set to True, even if the
                lattice is non-symmorphic, only symmorphic space group symmetry
                will be considered. Default is False, meaning the space group is
                determined by the lattice symmetry to be symmorphic or non-symmorphic.
        '''
        if h is not None: self.h = h
        if a is not None: self.a = a
        if mesh is not None: self.mesh = mesh
        if nimgs is not None: self.nimgs = nimgs
        if dimension is not None: self.dimension = dimension
        if fractional is not None: self.fractional = fractional
        if precision is not None: self.precision = precision
        if rcut is not None: self.rcut = rcut
        if ke_cutoff is not None: self.ke_cutoff = ke_cutoff
        if low_dim_ft_type is not None: self.low_dim_ft_type = low_dim_ft_type
        if use_loose_rcut is not None:
            self.use_loose_rcut = use_loose_rcut
        if use_particle_mesh_ewald is not None:
            self.use_particle_mesh_ewald = use_particle_mesh_ewald
        if space_group_symmetry is not None:
            self.space_group_symmetry = space_group_symmetry
        if symmorphic is not None:
            self.symmorphic = symmorphic

        if self.a is None:
            raise RuntimeError('Lattice parameters not specified')

        if self.dimension == 1 and self.low_dim_ft_type != 'inf_vacuum':
            raise RuntimeError('Uniform grids for dimension=1 not supported')

        _built = self._built
        mole.MoleBase.build(self, False, parse_arg, *args, **kwargs)

        exp_min = np.array([self.bas_exp(ib).min() for ib in range(self.nbas)])
        if self.exp_to_discard is None:
            if np.any(exp_min < 0.1):
                sys.stderr.write('''WARNING!
  Very diffused basis functions are found in the basis set. They may lead to severe
  linear dependence and numerical instability.  You can set  cell.exp_to_discard=0.1
  to remove the diffused Gaussians whose exponents are less than 0.1.\n\n''')
        elif np.any(exp_min < self.exp_to_discard):
            # Discard functions of small exponents in basis
            _basis = {}
            for symb, basis_now in self._basis.items():
                basis_add = []
                for b in basis_now:
                    l = b[0]
                    if isinstance(b[1], (int, np.integer)):
                        kappa = b[1]
                        b_coeff = np.array(b[2:])
                    else:
                        kappa = 0
                        b_coeff = np.array(b[1:])
                    es = b_coeff[:,0]
                    if np.any(es < self.exp_to_discard):
                        b_coeff = b_coeff[es>=self.exp_to_discard]
# contraction coefficients may be completely zero after removing one primitive
# basis. Removing the zero-coefficient basis.
                        b_coeff = b_coeff[:,np.all(b_coeff!=0, axis=0)]
                        if b_coeff.size > 0:
                            if kappa == 0:
                                basis_add.append([l] + b_coeff.tolist())
                            else:
                                basis_add.append([l,kappa] + b_coeff.tolist())
                    else:
                        basis_add.append(b)
                _basis[symb] = basis_add
            self._basis = _basis

            steep_shls = []
            nprim_drop = 0
            nctr_drop = 0
            _env = self._env.copy()
            for ib in range(len(self._bas)):
                l = self.bas_angular(ib)
                nprim = self.bas_nprim(ib)
                nc = self.bas_nctr(ib)
                es = self.bas_exp(ib)
                ptr = self._bas[ib,mole.PTR_COEFF]
                cs = self._env[ptr:ptr+nprim*nc].reshape(nc,nprim).T

                if np.any(es < self.exp_to_discard):
                    cs = cs[es>=self.exp_to_discard]
                    es = es[es>=self.exp_to_discard]
                    nprim_old, nc_old = nprim, nc

# contraction coefficients may be completely zero after removing one primitive
# basis. Removing the zero-coefficient basis.
                    cs = cs[:,np.all(cs!=0, axis=0)]
                    nprim, nc = cs.shape
                    self._bas[ib,mole.NPRIM_OF] = nprim
                    self._bas[ib,mole.NCTR_OF] = nc

                    nprim_drop = nprim_old - nprim + nprim_drop
                    nctr_drop = nc_old - nc + nctr_drop
                    if cs.size > 0:
                        pe = self._bas[ib,mole.PTR_EXP]
                        _env[pe:pe+nprim] = es
                        cs = mole._nomalize_contracted_ao(l, es, cs)
                        _env[ptr:ptr+nprim*nc] = cs.T.reshape(-1)
                if nprim > 0:
                    steep_shls.append(ib)
            self._env = _env
            self._bas = np.asarray(self._bas[steep_shls], order='C')
            logger.info(self, 'Discarded %d diffused primitive functions, '
                        '%d contracted functions', nprim_drop, nctr_drop)
            #logger.debug1(self, 'Old shells %s', steep_shls)

        if self.rcut is None or self._rcut_from_build:
            self._rcut = estimate_rcut(self, self.precision)
            self._rcut_from_build = True

        _a = self.lattice_vectors()
        if np.linalg.det(_a) < 0:
            sys.stderr.write('''WARNING!
  Lattice are not in right-handed coordinate system. This can cause wrong value for some integrals.
  It's recommended to resort the lattice vectors to\na = %s\n\n''' % _a[[0,2,1]])

        if self.dimension <= 2 and self.low_dim_ft_type != 'inf_vacuum':
            # check vacuum size. See Fig 1 of PRB, 73, 2015119
            #Lz_guess = self.rcut*(1+np.sqrt(2))
            Lz_guess = self.rcut * 2
            if np.linalg.norm(_a[2]) < 0.7 * Lz_guess:
                sys.stderr.write('''WARNING!
  Size of vacuum may not be enough. The recommended vacuum size is %s AA (%s Bohr)\n\n'''
                                 % (Lz_guess*param.BOHR, Lz_guess))

        if self.mesh is None or self._mesh_from_build:
            if self.ke_cutoff is None:
                ke_cutoff = estimate_ke_cutoff(self, self.precision)
            else:
                ke_cutoff = self.ke_cutoff
            self._mesh = pbctools.cutoff_to_mesh(_a, ke_cutoff)

            if self.dimension <= 2 and self.low_dim_ft_type == 'inf_vacuum':
                self._mesh[self.dimension:] = _mesh_inf_vaccum(self)
            self._mesh_from_build = True

        if self.space_group_symmetry:
            _check_mesh_symm = not self._mesh_from_build
            self.build_lattice_symmetry(check_mesh_symmetry=_check_mesh_symm)

        if dump_input and not _built and self.verbose > logger.NOTE:
            self.dump_input()
            logger.info(self, 'lattice vectors  a1 [%.9f, %.9f, %.9f]', *_a[0])
            logger.info(self, '                 a2 [%.9f, %.9f, %.9f]', *_a[1])
            logger.info(self, '                 a3 [%.9f, %.9f, %.9f]', *_a[2])
            logger.info(self, 'dimension = %s', self.dimension)
            logger.info(self, 'low_dim_ft_type = %s', self.low_dim_ft_type)
            logger.info(self, 'Cell volume = %g', self.vol)
            # Check atoms coordinates
            if self.dimension > 0 and self.natm > 0:
                scaled_atom_coords = self.get_scaled_atom_coords(_a)
                atom_boundary_max = scaled_atom_coords[:,:self.dimension].max(axis=0)
                atom_boundary_min = scaled_atom_coords[:,:self.dimension].min(axis=0)
                if (np.any(atom_boundary_max > 1) or np.any(atom_boundary_min < -1)):
                    logger.warn(self, 'Atoms found out of the primitive cell.')

            if self.exp_to_discard is not None:
                logger.info(self, 'exp_to_discard = %s', self.exp_to_discard)
            logger.info(self, 'rcut = %s (nimgs = %s)', self.rcut, self.nimgs)
            logger.info(self, 'lattice sum = %d cells', len(self.get_lattice_Ls()))
            logger.info(self, 'precision = %g', self.precision)
            logger.info(self, 'pseudo = %s', self.pseudo)
            if ke_cutoff is not None:
                logger.info(self, 'ke_cutoff = %s', ke_cutoff)
                logger.info(self, '    = %s mesh (%d PWs)',
                            self.mesh, np.prod(self.mesh))
            else:
                logger.info(self, 'mesh = %s (%d PWs)',
                            self.mesh, np.prod(self.mesh))
                Ecut = pbctools.mesh_to_cutoff(self.lattice_vectors(), self.mesh)
                logger.info(self, '    = ke_cutoff %s', Ecut)
            if self.space_group_symmetry:
                self.lattice_symmetry.dump_info()
        return self
    kernel = build

    @property
    def h(self):
        return np.asarray(self.a).T
    @h.setter
    def h(self, x):
        warnings.warn('cell.h is deprecated.  It is replaced by the '
                      '(row-based) lattice vectors cell.a:  cell.a = cell.h.T\n')
        if isinstance(x, str):
            x = x.replace(';',' ').replace(',',' ').replace('\n',' ')
            self.a = np.asarray([float(z) for z in x.split()]).reshape(3,3).T
        else:
            self.a = np.asarray(x).T

    @property
    def _h(self):
        return self.lattice_vectors().T

    @property
    def vol(self):
        return abs(np.linalg.det(self.lattice_vectors()))

    @property
    def Gv(self):
        return self.get_Gv()

    @property
    def gs(self):
        return [n//2 for n in self.mesh]
    @gs.setter
    def gs(self, x):
        warnings.warn('cell.gs is deprecated.  It is replaced by cell.mesh,'
                      'the number of PWs (=2*gs+1) along each direction.')
        self.mesh = [2*n+1 for n in x]

    @property
    def drop_exponent(self):
        return self.exp_to_discard
    @drop_exponent.setter
    def drop_exponent(self, x):
        self.exp_to_discard = x

    @property
    def nimgs(self):
        return self.get_bounding_sphere(self.rcut)
    @nimgs.setter
    def nimgs(self, x):
        b = self.reciprocal_vectors(norm_to=1)
        heights_inv = lib.norm(b, axis=1)
        self.rcut = max(np.asarray(x) / heights_inv)
        rcut_guess = estimate_rcut(self, self.precision)
        if self.rcut > rcut_guess*1.5:
            msg = ('.nimgs is a deprecated attribute.  It is replaced by .rcut '
                   'attribute for lattice sum cutoff radius.  The given nimgs '
                   '%s is far over the estimated cutoff radius %s. ' %
                   (x, rcut_guess))
            warnings.warn(msg)

    def lattice_vectors(self):
        '''Convert the primitive lattice vectors.

        Return 3x3 array in which each row represents one direction of the
        lattice vectors (unit in Bohr)
        '''
        if isinstance(self.a, str):
            a = self.a.replace(';',' ').replace(',',' ').replace('\n',' ')
            a = np.asarray([float(x) for x in a.split()]).reshape(3,3)
        else:
            a = np.asarray(self.a, dtype=np.double).reshape(3,3)
        if isinstance(self.unit, str):
            if is_au(self.unit):
                return a
            else:
                return a/param.BOHR
        else:
            return a/self.unit

    def get_scaled_atom_coords(self, a=None):
        ''' Get scaled atomic coordinates.
        '''
        if a is None:
            a = self.lattice_vectors()
        return np.dot(self.atom_coords(), np.linalg.inv(a))

    def reciprocal_vectors(self, norm_to=2*np.pi):
        r'''
        .. math::

            \begin{align}
            \mathbf{b_1} &= 2\pi \frac{\mathbf{a_2} \times \mathbf{a_3}}{\mathbf{a_1} \cdot (\mathbf{a_2} \times \mathbf{a_3})} \\
            \mathbf{b_2} &= 2\pi \frac{\mathbf{a_3} \times \mathbf{a_1}}{\mathbf{a_2} \cdot (\mathbf{a_3} \times \mathbf{a_1})} \\
            \mathbf{b_3} &= 2\pi \frac{\mathbf{a_1} \times \mathbf{a_2}}{\mathbf{a_3} \cdot (\mathbf{a_1} \times \mathbf{a_2})}
            \end{align}

        '''  # noqa: E501
        a = self.lattice_vectors()
        if self.dimension == 1:
            assert (abs(np.dot(a[0], a[1])) < 1e-9 and
                   abs(np.dot(a[0], a[2])) < 1e-9 and
                   abs(np.dot(a[1], a[2])) < 1e-9)
        elif self.dimension == 2:
            assert (abs(np.dot(a[0], a[2])) < 1e-9 and
                   abs(np.dot(a[1], a[2])) < 1e-9)
        b = np.linalg.inv(a.T)
        return norm_to * b

    def get_abs_kpts(self, scaled_kpts):
        '''Get absolute k-points (in 1/Bohr), given "scaled" k-points in
        fractions of lattice vectors.

        Args:
            scaled_kpts : (nkpts, 3) ndarray of floats

        Returns:
            abs_kpts : (nkpts, 3) ndarray of floats
        '''
        return np.dot(scaled_kpts, self.reciprocal_vectors())

    def get_scaled_kpts(self, abs_kpts, kpts_in_ibz=True):
        '''Get scaled k-points, given absolute k-points in 1/Bohr.

        Args:
            abs_kpts : (nkpts, 3) ndarray of floats or :class:`KPoints` object

            kpts_in_ibz : bool
                If True, return k-points in IBZ; otherwise, return k-points in BZ.
                Default value is True. This has effects only if abs_kpts is a
                :class:`KPoints` object

        Returns:
            scaled_kpts : (nkpts, 3) ndarray of floats
        '''
        from pyscf.pbc.lib.kpts import KPoints
        if isinstance(abs_kpts, KPoints):
            if kpts_in_ibz:
                return abs_kpts.kpts_scaled_ibz
            else:
                return abs_kpts.kpts_scaled
        return 1./(2*np.pi)*np.dot(abs_kpts, self.lattice_vectors().T)

    def cutoff_to_mesh(self, ke_cutoff):
        '''Convert KE cutoff to FFT-mesh

        Args:
            ke_cutoff : float
                KE energy cutoff in a.u.

        Returns:
            mesh : (3,) array
        '''
        a = self.lattice_vectors()
        dim = self.dimension
        mesh = pbctools.cutoff_to_mesh(a, ke_cutoff)
        if dim < 2 or (dim == 2 and self.low_dim_ft_type == 'inf_vacuum'):
            mesh[dim:] = self.mesh[dim:]
        return mesh

    make_kpts = get_kpts = make_kpts

    pack = pack

    @classmethod
    @lib.with_doc(unpack.__doc__)
    def unpack(cls, moldic):
        return unpack(moldic)

    @lib.with_doc(unpack.__doc__)
    def unpack_(self, moldic):
        self.__dict__.update(moldic)
        return self

    dumps = dumps

    @classmethod
    @lib.with_doc(loads.__doc__)
    def loads(cls, molstr):
        return loads(molstr)

    @lib.with_doc(unpack.__doc__)
    def loads_(self, molstr):
        self.__dict__.update(loads(molstr).__dict__)
        return self

    bas_rcut = bas_rcut
    rcut_by_shells = rcut_by_shells

    get_lattice_Ls = pbctools.get_lattice_Ls

    get_nimgs = get_nimgs

    get_ewald_params = get_ewald_params

    get_bounding_sphere = get_bounding_sphere

    get_Gv = get_Gv
    get_Gv_weights = get_Gv_weights

    get_SI = get_SI

    ewald = ewald
    energy_nuc = get_enuc = ewald

    gen_uniform_grids = get_uniform_grids = get_uniform_grids

    __add__ = conc_cell

    def pbc_intor(self, intor, comp=None, hermi=0, kpts=None, kpt=None,
                  shls_slice=None, **kwargs):
        r'''One-electron integrals with PBC.

        .. math::

            \sum_T \int \mu(r) * [intor] * \nu (r-T) dr

        See also Mole.intor
        '''
        if not self._built:
            logger.warn(self, 'Warning: intor envs of %s not initialized.', self)
            # FIXME: Whether to check _built and call build?  ._bas and .basis
            # may not be consistent. calling .build() may leads to wrong intor env.
            #self.build(False, False)
        if self.use_loose_rcut:
            return _intor_cross_screened(
                            intor, self, self, comp, hermi, kpts, kpt,
                            shls_slice, **kwargs)
        return intor_cross(intor, self, self, comp, hermi, kpts, kpt,
                           shls_slice, **kwargs)

    pbc_eval_ao = pbc_eval_gto = pbc_eval_gto

    @lib.with_doc(pbc_eval_gto.__doc__)
    def eval_gto(self, eval_name, coords, comp=None, kpts=None, kpt=None,
                 shls_slice=None, non0tab=None, ao_loc=None, cutoff=None,
                 out=None):
        if eval_name[:3] == 'PBC':
            return self.pbc_eval_gto(eval_name, coords, comp, kpts, kpt,
                                     shls_slice, non0tab, ao_loc, cutoff, out)
        else:
            return mole.eval_gto(self, eval_name, coords, comp,
                                 shls_slice, non0tab, ao_loc, cutoff, out)
    eval_ao = eval_gto

    def from_ase(self, ase_atom):
        '''Update cell based on given ase atom object

        Examples:

        >>> from ase.lattice import bulk
        >>> cell.from_ase(bulk('C', 'diamond', a=LATTICE_CONST))
        '''
        from pyscf.pbc.tools import pyscf_ase
        self.a = ase_atom.cell
        self.atom = pyscf_ase.ase_atoms_to_pyscf(ase_atom)
        return self

    def to_mol(self):
        '''Return a Mole object using the same atoms and basis functions as
        the Cell object.
        '''
        mol = self.view(mole.Mole)
        del mol.a
        mol.__dict__.pop('fractional', None)
        mol.__dict__.pop('ke_cutoff', None)
        mol.__dict__.pop('_mesh', None)
        mol.__dict__.pop('_rcut', None)
        mol.__dict__.pop('dimension', None)
        mol.__dict__.pop('low_dim_ft_type', None)
        mol.enuc = None #reset nuclear energy
        if mol.symmetry:
            mol._build_symmetry()
        return mol

del (INTEGRAL_PRECISION, WRAP_AROUND, WITH_GAMMA, EXP_DELIMITER)



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Short range part of ECP under PBC
'''

from functools import reduce
import numpy as np
from pyscf import lib
from pyscf import gto
from pyscf.pbc.gto.cell import _split_basis
from pyscf.pbc.df import incore


def ecp_int(cell, kpts=None):
    lib.logger.debug(cell, 'PBC-ECP integrals')
    if kpts is None:
        kpts_lst = np.zeros((1,3))
    else:
        kpts_lst = np.reshape(kpts, (-1,3))

    cell, contr_coeff = _split_basis(cell)
    lib.logger.debug1(cell, 'nao %d -> nao %d', *(contr_coeff.shape))

    ecpcell = cell.copy(deep=False)
    # append a fake s function to mimic the auxiliary index in pbc.incore.
    exp_ptr = cell._ecpbas[-1,gto.PTR_EXP]
    ecpcell._bas = np.array([[0, 0, 1, 1, 0, exp_ptr, 0, 0]], dtype=np.int32)
    # _env[AS_ECPBAS_OFFSET] is to be determined in pbc.incore
    cell._env[gto.AS_NECPBAS] = len(cell._ecpbas)
    # shls_slice of auxiliary index (0,1) corresponds to the fake s function
    shls_slice = (0, cell.nbas, 0, cell.nbas, 0, 1)

    dfbuilder = incore.Int3cBuilder(cell, ecpcell, kpts_lst).build()
    int3c = dfbuilder.gen_int3c_kernel('ECPscalar', aosym='s2', comp=1,
                                       j_only=True, return_complex=True)
    buf = int3c(shls_slice)
    buf = buf.reshape(len(kpts_lst),-1)
    mat = []
    for k, kpt in enumerate(kpts_lst):
        v = lib.unpack_tril(buf[k], lib.HERMITIAN)
        if abs(kpt).max() < 1e-9:  # gamma_point:
            v = v.real
        mat.append(reduce(np.dot, (contr_coeff.T, v, contr_coeff)))
    if kpts is None or np.shape(kpts) == (3,):
        mat = mat[0]
    return mat



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Qiming Sun <osirpt.sun@gmail.com>
#          Junzi Liu <latrix1247@gmail.com>
#          Susi Lehtola <susi.lehtola@gmail.com>

from functools import reduce
import numpy
import scipy.linalg
from pyscf import lib
from pyscf.gto import mole
from pyscf.lib import logger
from pyscf.lib.scipy_helper import pivoted_cholesky
from pyscf import __config__

LINEAR_DEP_THRESHOLD = getattr(__config__, 'scf_addons_remove_linear_dep_threshold', 1e-8)
CHOLESKY_THRESHOLD = getattr(__config__, 'scf_addons_cholesky_threshold', 1e-10)
FORCE_PIVOTED_CHOLESKY = getattr(__config__, 'scf_addons_force_cholesky', False)
LINEAR_DEP_TRIGGER = getattr(__config__, 'scf_addons_remove_linear_dep_trigger', 1e-10)

SMEARING_METHOD = getattr(__config__, 'pbc_scf_addons_smearing_method', 'fermi')

def smearing(mf, sigma=None, method=SMEARING_METHOD, mu0=None, fix_spin=False):
    '''Fermi-Dirac or Gaussian smearing'''
    if isinstance(mf, _SmearingSCF):
        mf.sigma = sigma
        mf.method = method
        mf.mu0 = mu0
        mf.fix_spin = fix_spin
        return mf

    assert not mf.istype('KSCF')
    if mf.istype('ROHF'):
        # Roothaan Fock matrix does not make much sense for smearing.
        # Restore the conventional RHF treatment.
        from pyscf import scf
        from pyscf import dft
        known_class = {
            dft.rks_symm.ROKS: dft.rks_symm.RKS,
            dft.roks.ROKS    : dft.rks.RKS     ,
            scf.hf_symm.ROHF : scf.hf_symm.RHF ,
            scf.rohf.ROHF    : scf.hf.RHF      ,
        }
        mf = _object_without_soscf(mf, known_class)
    return lib.set_class(_SmearingSCF(mf, sigma, method, mu0, fix_spin),
                         (_SmearingSCF, mf.__class__))

def smearing_(mf, *args, **kwargs):
    mf1 = smearing(mf, *args, **kwargs)
    mf.__class__ = mf1.__class__
    mf.__dict__ = mf1.__dict__
    return mf

def _get_grad_tril(mo_coeff, mo_occ, fock):
    f_mo = reduce(numpy.dot, (mo_coeff.T.conj(), fock, mo_coeff))
    return f_mo[numpy.tril_indices_from(f_mo, -1)]

def _fermi_smearing_occ(mu, mo_energy, sigma):
    '''Fermi-Dirac smearing'''
    occ = numpy.zeros_like(mo_energy)
    de = (mo_energy - mu) / sigma
    occ[de<40] = 1./(numpy.exp(de[de<40])+1.)
    return occ

def _gaussian_smearing_occ(mu, mo_energy, sigma):
    '''Gaussian smearing'''
    return 0.5 * scipy.special.erfc((mo_energy - mu) / sigma)

def _smearing_optimize(f_occ, mo_es, nocc, sigma):
    def nelec_cost_fn(m):
        mo_occ = f_occ(m, mo_es, sigma)
        return (mo_occ.sum() - nocc)**2

    mu0 = _init_guess_mu(f_occ, mo_es, nocc, sigma)
    res = scipy.optimize.minimize(
        nelec_cost_fn, mu0, method='Powell',
        options={'xtol': 1e-5, 'ftol': 1e-5, 'maxiter': 10000})
    mu = res.x
    mo_occs = f_occ(mu, mo_es, sigma)
    return mu, mo_occs

def _init_guess_mu(f_occ, mo_es, nocc, sigma, box=[-1.,1.], step=1e-3):
    ''' Generate an initial guess for `mu` by manually minimizing the nocc error on a grid
        around `fermi`. The grid is constructed as:

            mus = fermi + numpy.arange(*box, step) / 27.211399

        By default, `box = [-1, 1]` and `step = 1e-3`, which corresponds to searching
        within a +/-1 eV window around `fermi` with a resolution of 1 meV (resulting
        in 2000 grid points).

        Return:
            mu0 : float
                Best initial guess within the searching window
    '''
    fermi = _get_fermi(mo_es, nocc)
    mus = fermi + numpy.arange(*box, step)/27.211399
    noccs = numpy.asarray([f_occ(m, mo_es, sigma).sum() for m in mus])
    return mus[abs(noccs-nocc).argmin()]

def _get_fermi(mo_energy, nocc):
    mo_e_sorted = numpy.sort(mo_energy)
    if isinstance(nocc, (int, numpy.integer)):
        return mo_e_sorted[nocc-1]
    else: # nocc = ?.5 or nocc = ?.0
        return mo_e_sorted[numpy.ceil(nocc).astype(int) - 1]

class _SmearingSCF:

    __name_mixin__ = 'Smearing'

    _keys = {
        'sigma', 'smearing_method', 'mu0', 'fix_spin', 'entropy', 'e_free', 'e_zero'
    }

    def __init__(self, mf, sigma, method, mu0, fix_spin):
        self.__dict__.update(mf.__dict__)
        self.sigma = sigma
        self.smearing_method = method
        self.mu0 = mu0
        self.fix_spin = fix_spin
        self.entropy = None
        self.e_free = None
        self.e_zero = None

    def undo_smearing(self):
        obj = lib.view(self, lib.drop_class(self.__class__, _SmearingSCF))
        del obj.sigma
        del obj.smearing_method
        del obj.fix_spin
        del obj.entropy
        del obj.e_free
        del obj.e_zero
        return obj

    def get_occ(self, mo_energy=None, mo_coeff=None):
        '''Label the occupancies for each orbital
        '''
        from pyscf import scf
        from pyscf.pbc.tools import print_mo_energy_occ
        if (self.sigma == 0) or (not self.sigma) or (not self.smearing_method):
            mo_occ = super().get_occ(mo_energy, mo_coeff)
            return mo_occ

        is_uhf = self.istype('UHF')
        is_rhf = self.istype('RHF')
        if isinstance(self, scf.rohf.ROHF):
            # ROHF leads to two Fock matrices. It's not clear how to define the
            # Roothaan effective Fock matrix from the two.
            raise NotImplementedError('Smearing-ROHF')

        sigma = self.sigma
        if self.smearing_method.lower() == 'fermi':
            f_occ = _fermi_smearing_occ
        else:
            f_occ = _gaussian_smearing_occ

        if self.fix_spin and is_uhf: # spin separated fermi level
            mo_es = mo_energy
            nocc = self.nelec
            if self.mu0 is None:
                mu_a, occa = _smearing_optimize(f_occ, mo_es[0], nocc[0], sigma)
                mu_b, occb = _smearing_optimize(f_occ, mo_es[1], nocc[1], sigma)
            else:
                if numpy.isscalar(self.mu0):
                    mu_a = mu_b = self.mu0
                elif len(self.mu0) == 2:
                    mu_a, mu_b = self.mu0
                else:
                    raise TypeError(f'Unsupported mu0: {self.mu0}')
                occa = f_occ(mu_a, mo_es[0], sigma)
                occb = f_occ(mu_b, mo_es[1], sigma)
            mu = [mu_a, mu_b]
            mo_occs = [occa, occb]
            self.entropy  = self._get_entropy(mo_es[0], mo_occs[0], mu[0])
            self.entropy += self._get_entropy(mo_es[1], mo_occs[1], mu[1])
            fermi = (_get_fermi(mo_es[0], nocc[0]), _get_fermi(mo_es[1], nocc[1]))

            logger.debug(self, '    Alpha-spin Fermi level %g  Sum mo_occ = %s  should equal nelec = %s',
                         fermi[0], mo_occs[0].sum(), nocc[0])
            logger.debug(self, '    Beta-spin  Fermi level %g  Sum mo_occ = %s  should equal nelec = %s',
                         fermi[1], mo_occs[1].sum(), nocc[1])
            logger.info(self, '    sigma = %g  Optimized mu_alpha = %.12g  entropy = %.12g',
                        sigma, mu[0], self.entropy)
            logger.info(self, '    sigma = %g  Optimized mu_beta  = %.12g  entropy = %.12g',
                        sigma, mu[1], self.entropy)
            if self.verbose >= logger.DEBUG:
                print_mo_energy_occ(self, mo_energy, mo_occs, True)
        else: # all orbitals treated with the same fermi level
            nelectron = self.mol.nelectron
            if is_uhf:
                mo_es = numpy.hstack(mo_energy)
            else:
                mo_es = mo_energy
            if is_rhf:
                nelectron = nelectron / 2

            if self.mu0 is None:
                mu, mo_occs = _smearing_optimize(f_occ, mo_es, nelectron, sigma)
            else:
                # If mu0 is given, fix mu instead of electron number. XXX -Chong Sun
                mu = self.mu0
                assert numpy.isscalar(mu)
                mo_occs = f_occ(mu, mo_es, sigma)
            self.entropy = self._get_entropy(mo_es, mo_occs, mu)
            if is_rhf:
                mo_occs *= 2
                self.entropy *= 2

            fermi = _get_fermi(mo_es, nelectron)
            logger.debug(self, '    Fermi level %g  Sum mo_occ = %s  should equal nelec = %s',
                         fermi, mo_occs.sum(), nelectron)
            logger.info(self, '    sigma = %g  Optimized mu = %.12g  entropy = %.12g',
                        sigma, mu, self.entropy)
            if is_uhf:
                mo_occs = mo_occs.reshape(2, -1)
            if self.verbose >= logger.DEBUG:
                print_mo_energy_occ(self, mo_energy, mo_occs, is_uhf)
        return mo_occs

    # See https://www.vasp.at/vasp-workshop/slides/k-points.pdf
    def _get_entropy(self, mo_energy, mo_occ, mu):
        if self.smearing_method.lower() == 'fermi':
            f = mo_occ
            f = f[(f>0) & (f<1)]
            entropy = -(f*numpy.log(f) + (1-f)*numpy.log(1-f)).sum()
        else:
            entropy = (numpy.exp(-((mo_energy-mu)/self.sigma)**2).sum()
                       / (2*numpy.sqrt(numpy.pi)))
        return entropy

    def get_grad(self, mo_coeff, mo_occ, fock=None):
        if (self.sigma == 0) or (not self.sigma) or (not self.smearing_method):
            return super().get_grad(mo_coeff, mo_occ, fock)

        if fock is None:
            dm1 = self.make_rdm1(mo_coeff, mo_occ)
            fock = self.get_hcore() + self.get_veff(self.mol, dm1)
        if self.istype('UHF'):
            ga = _get_grad_tril(mo_coeff[0], mo_occ[0], fock[0])
            gb = _get_grad_tril(mo_coeff[1], mo_occ[1], fock[1])
            return numpy.hstack((ga,gb))
        else: # rhf and ghf
            return _get_grad_tril(mo_coeff, mo_occ, fock)

    def energy_tot(self, dm=None, h1e=None, vhf=None):
        e_tot = self.energy_elec(dm, h1e, vhf)[0] + self.energy_nuc()
        if self.sigma and self.smearing_method and self.entropy is not None:
            self.e_free = e_tot - self.sigma * self.entropy
            self.e_zero = e_tot - self.sigma * self.entropy * .5
            logger.info(self, '    Total E(T) = %.15g  Free energy = %.15g  E0 = %.15g',
                        e_tot, self.e_free, self.e_zero)
        return e_tot

def frac_occ_(mf, tol=1e-3):
    '''
    Addons for SCF methods to assign fractional occupancy for degenerated
    occupied HOMOs.

    Examples::

        >>> mf = gto.M(atom='O 0 0 0; O 0 0 1', verbose=4).RHF()
        >>> mf = scf.addons.frac_occ(mf)
        >>> mf.run()
    '''
    from pyscf.scf import uhf, rohf
    old_get_occ = mf.get_occ
    mol = mf.mol

    def guess_occ(mo_energy, nocc):
        mo_occ = numpy.zeros_like(mo_energy)
        if nocc:
            sorted_idx = numpy.argsort(mo_energy)
            homo = mo_energy[sorted_idx[nocc-1]]
            lumo = mo_energy[sorted_idx[nocc]]
            frac_occ_lst = abs(mo_energy - homo) < tol
            integer_occ_lst = (mo_energy <= homo) & (~frac_occ_lst)
            mo_occ[integer_occ_lst] = 1
            degen = numpy.count_nonzero(frac_occ_lst)
            frac = nocc - numpy.count_nonzero(integer_occ_lst)
            mo_occ[frac_occ_lst] = float(frac) / degen
        else:
            homo = 0.0
            lumo = 0.0
            frac_occ_lst = numpy.zeros_like(mo_energy, dtype=bool)
        return mo_occ, numpy.where(frac_occ_lst)[0], homo, lumo

    if mf.istype('UHF'):
        def get_occ(mo_energy, mo_coeff=None):
            nocca, noccb = mol.nelec
            mo_occa, frac_lsta, homoa, lumoa = guess_occ(mo_energy[0], nocca)
            mo_occb, frac_lstb, homob, lumob = guess_occ(mo_energy[1], noccb)

            if abs(homoa - lumoa) < tol or abs(homob - lumob) < tol:
                mo_occ = numpy.array([mo_occa, mo_occb])
                if len(frac_lstb):
                    logger.warn(mf, 'fraction occ = %6g for alpha orbitals %s  '
                                '%6g for beta orbitals %s',
                                mo_occa[frac_lsta[0]], frac_lsta,
                                mo_occb[frac_lstb[0]], frac_lstb)
                    logger.info(mf, '  alpha HOMO = %.12g  LUMO = %.12g', homoa, lumoa)
                    logger.info(mf, '  beta  HOMO = %.12g  LUMO = %.12g', homob, lumob)
                    logger.debug(mf, '  alpha mo_energy = %s', mo_energy[0])
                    logger.debug(mf, '  beta  mo_energy = %s', mo_energy[1])
                else:
                    logger.warn(mf, 'fraction occ = %6g for alpha orbitals %s  ',
                                mo_occa[frac_lsta[0]], frac_lsta)
                    logger.info(mf, '  alpha HOMO = %.12g  LUMO = %.12g', homoa, lumoa)
                    logger.debug(mf, '  alpha mo_energy = %s', mo_energy[0])
            else:
                mo_occ = old_get_occ(mo_energy, mo_coeff)
            return mo_occ

    elif mf.istype('ROHF'):
        def get_occ(mo_energy, mo_coeff=None):
            nocca, noccb = mol.nelec
            mo_occa, frac_lsta, homoa, lumoa = guess_occ(mo_energy, nocca)
            mo_occb, frac_lstb, homob, lumob = guess_occ(mo_energy, noccb)

            if abs(homoa - lumoa) < tol or abs(homob - lumob) < tol:
                mo_occ = mo_occa + mo_occb
                if len(frac_lstb):
                    logger.warn(mf, 'fraction occ = %6g for alpha orbitals %s  '
                                '%6g for beta orbitals %s',
                                mo_occa[frac_lsta[0]], frac_lsta,
                                mo_occb[frac_lstb[0]], frac_lstb)
                    logger.info(mf, '  HOMO = %.12g  LUMO = %.12g', homoa, lumoa)
                    logger.debug(mf, '  mo_energy = %s', mo_energy)
                else:
                    logger.warn(mf, 'fraction occ = %6g for alpha orbitals %s  ',
                                mo_occa[frac_lsta[0]], frac_lsta)
                    logger.info(mf, '  HOMO = %.12g  LUMO = %.12g', homoa, lumoa)
                    logger.debug(mf, '  mo_energy = %s', mo_energy)
            else:
                mo_occ = old_get_occ(mo_energy, mo_coeff)
            return mo_occ

        def get_grad(mo_coeff, mo_occ, fock):
            occidxa = mo_occ > 0
            occidxb = mo_occ > 1
            viridxa = ~occidxa
            viridxb = ~occidxb
            uniq_var_a = viridxa.reshape(-1,1) & occidxa
            uniq_var_b = viridxb.reshape(-1,1) & occidxb

            if getattr(fock, 'focka', None) is not None:
                focka = fock.focka
                fockb = fock.fockb
            elif getattr(fock, 'ndim', None) == 3:
                focka, fockb = fock
            else:
                focka = fockb = fock
            focka = reduce(numpy.dot, (mo_coeff.T.conj(), focka, mo_coeff))
            fockb = reduce(numpy.dot, (mo_coeff.T.conj(), fockb, mo_coeff))

            g = numpy.zeros_like(focka)
            g[uniq_var_a]  = focka[uniq_var_a]
            g[uniq_var_b] += fockb[uniq_var_b]
            return g[uniq_var_a | uniq_var_b]

    elif mf.istype('RHF'):
        def get_occ(mo_energy, mo_coeff=None):
            nocc = (mol.nelectron+1) // 2  # n_docc + n_socc
            mo_occ, frac_lst, homo, lumo = guess_occ(mo_energy, nocc)
            n_docc = mol.nelectron // 2
            n_socc = nocc - n_docc
            if abs(homo - lumo) < tol or n_socc:
                mo_occ *= 2
                degen = len(frac_lst)
                mo_occ[frac_lst] -= float(n_socc) / degen
                logger.warn(mf, 'fraction occ = %6g  for orbitals %s',
                            mo_occ[frac_lst[0]], frac_lst)
                logger.info(mf, 'HOMO = %.12g  LUMO = %.12g', homo, lumo)
                logger.debug(mf, '  mo_energy = %s', mo_energy)
            else:
                mo_occ = old_get_occ(mo_energy, mo_coeff)
            return mo_occ
    else:
        def get_occ(mo_energy, mo_coeff=None):
            nocc = mol.nelectron
            mo_occ, frac_lst, homo, lumo = guess_occ(mo_energy, nocc)
            if abs(homo - lumo) < tol:
                logger.warn(mf, 'fraction occ = %6g  for orbitals %s',
                            mo_occ[frac_lst[0]], frac_lst)
                logger.info(mf, 'HOMO = %.12g  LUMO = %.12g', homo, lumo)
                logger.debug(mf, '  mo_energy = %s', mo_energy)
            else:
                mo_occ = old_get_occ(mo_energy, mo_coeff)
            return mo_occ

    mf.get_occ = get_occ
    try:
        mf.get_grad = get_grad
    except NameError:
        pass
    return mf
frac_occ = frac_occ_


def dynamic_occ_(mf, tol=1e-3):
    '''
    Dynamically adjust the occupancy to avoid degeneracy between HOMO and LUMO
    '''
    assert mf.istype('RHF')
    old_get_occ = mf.get_occ
    def get_occ(mo_energy, mo_coeff=None):
        mol = mf.mol
        nocc = mol.nelectron // 2
        sort_mo_energy = numpy.sort(mo_energy)
        lumo = sort_mo_energy[nocc]
        if abs(sort_mo_energy[nocc-1] - lumo) < tol:
            mo_occ = numpy.zeros_like(mo_energy)
            mo_occ[mo_energy<lumo] = 2
            lst = abs(mo_energy - lumo) < tol
            mo_occ[lst] = 0
            logger.warn(mf, 'set charge = %d', mol.charge+int(lst.sum())*2)
            logger.info(mf, 'HOMO = %.12g  LUMO = %.12g',
                        sort_mo_energy[nocc-1], sort_mo_energy[nocc])
            logger.debug(mf, '  mo_energy = %s', sort_mo_energy)
        else:
            mo_occ = old_get_occ(mo_energy, mo_coeff)
        return mo_occ
    mf.get_occ = get_occ
    return mf
dynamic_occ = dynamic_occ_

def dynamic_level_shift_(mf, factor=1.):
    '''Dynamically change the level shift in each SCF cycle.  The level shift
    value is set to (HF energy change * factor)
    '''
    old_get_fock = mf.get_fock
    mf._last_e = None
    def get_fock(h1e, s1e, vhf, dm, cycle=-1, diis=None,
                 diis_start_cycle=None, level_shift_factor=None, damp_factor=None,
                 fock_last=None):
        if cycle > 0 or diis is not None:
            if 'exc' in mf.scf_summary:  # DFT
                e_tot = mf.scf_summary['e1'] + mf.scf_summary['coul'] + mf.scf_summary['exc']
            else:
                e_tot = mf.scf_summary['e1'] + mf.scf_summary['e2']
            if mf._last_e is not None:
                level_shift_factor = abs(e_tot-mf._last_e) * factor
                logger.info(mf, 'Set level shift to %g', level_shift_factor)
            mf._last_e = e_tot
        return old_get_fock(h1e, s1e, vhf, dm, cycle, diis, diis_start_cycle,
                            level_shift_factor, damp_factor, fock_last=fock_last)
    mf.get_fock = get_fock
    return mf
dynamic_level_shift = dynamic_level_shift_

def float_occ_(mf):
    '''
    For UHF, allowing the Sz value being changed during SCF iteration.
    Determine occupation of alpha and beta electrons based on energy spectrum
    '''
    from pyscf.scf import uhf
    assert mf.istype('UHF')
    def get_occ(mo_energy, mo_coeff=None):
        mol = mf.mol
        ee = numpy.sort(numpy.hstack(mo_energy))
        n_a = numpy.count_nonzero(mo_energy[0]<(ee[mol.nelectron-1]+1e-3))
        n_b = mol.nelectron - n_a
        if mf.nelec is None:
            nelec = mf.mol.nelec
        else:
            nelec = mf.nelec
        if n_a != nelec[0]:
            logger.info(mf, 'change num. alpha/beta electrons '
                        ' %d / %d -> %d / %d',
                        nelec[0], nelec[1], n_a, n_b)
            mf.nelec = (n_a, n_b)
        return uhf.UHF.get_occ(mf, mo_energy, mo_coeff)
    mf.get_occ = get_occ
    return mf
dynamic_sz_ = float_occ = float_occ_

def follow_state_(mf, occorb=None):
    occstat = [occorb]
    old_get_occ = mf.get_occ
    def get_occ(mo_energy, mo_coeff=None):
        if occstat[0] is None:
            mo_occ = old_get_occ(mo_energy, mo_coeff)
        else:
            mo_occ = numpy.zeros_like(mo_energy)
            s = reduce(numpy.dot, (occstat[0].T, mf.get_ovlp(), mo_coeff))
            nocc = mf.mol.nelectron // 2
            #choose a subset of mo_coeff, which maximizes <old|now>
            idx = numpy.argsort(numpy.einsum('ij,ij->j', s, s))
            mo_occ[idx[-nocc:]] = 2
            logger.debug(mf, '  mo_occ = %s', mo_occ)
            logger.debug(mf, '  mo_energy = %s', mo_energy)
        occstat[0] = mo_coeff[:,mo_occ>0]
        return mo_occ
    mf.get_occ = get_occ
    return mf
follow_state = follow_state_

def mom_occ_(mf, occorb, setocc):
    '''Use maximum overlap method to determine occupation number for each orbital in every
    iteration.'''
    from pyscf.scf import uhf, rohf
    log = logger.Logger(mf.stdout, mf.verbose)
    if mf.istype('UHF'):
        coef_occ_a = occorb[0][:, setocc[0]>0]
        coef_occ_b = occorb[1][:, setocc[1]>0]
    elif mf.istype('ROHF'):
        if mf.mol.spin != (numpy.sum(setocc[0]) - numpy.sum(setocc[1])):
            raise ValueError('Wrong occupation setting for restricted open-shell calculation.')
        coef_occ_a = occorb[:, setocc[0]>0]
        coef_occ_b = occorb[:, setocc[1]>0]
    else: # GHF, and DHF
        assert setocc.ndim == 1

    if mf.istype('UHF') or mf.istype('ROHF'):
        def get_occ(mo_energy=None, mo_coeff=None):
            if mo_energy is None: mo_energy = mf.mo_energy
            if mo_coeff is None: mo_coeff = mf.mo_coeff
            if mf.istype('ROHF'):
                mo_coeff = numpy.array([mo_coeff, mo_coeff])
            mo_occ = numpy.zeros_like(setocc)
            nocc_a = int(numpy.sum(setocc[0]))
            nocc_b = int(numpy.sum(setocc[1]))
            s_a = reduce(numpy.dot, (coef_occ_a.conj().T, mf.get_ovlp(), mo_coeff[0]))
            s_b = reduce(numpy.dot, (coef_occ_b.conj().T, mf.get_ovlp(), mo_coeff[1]))
            #choose a subset of mo_coeff, which maximizes <old|now>
            idx_a = numpy.argsort(numpy.einsum('ij,ij->j', s_a, s_a))[::-1]
            idx_b = numpy.argsort(numpy.einsum('ij,ij->j', s_b, s_b))[::-1]
            mo_occ[0,idx_a[:nocc_a]] = 1.
            mo_occ[1,idx_b[:nocc_b]] = 1.

            log.debug(' New alpha occ pattern: %s', mo_occ[0])
            log.debug(' New beta occ pattern: %s', mo_occ[1])
            if isinstance(mf.mo_energy, numpy.ndarray) and mf.mo_energy.ndim == 1:
                log.debug1(' Current mo_energy(sorted) = %s', mo_energy)
            else:
                log.debug1(' Current alpha mo_energy(sorted) = %s', mo_energy[0])
                log.debug1(' Current beta mo_energy(sorted) = %s', mo_energy[1])

            if (int(numpy.sum(mo_occ[0])) != nocc_a):
                log.error('mom alpha electron occupation numbers do not match: %d, %d',
                          nocc_a, int(numpy.sum(mo_occ[0])))
            if (int(numpy.sum(mo_occ[1])) != nocc_b):
                log.error('mom beta electron occupation numbers do not match: %d, %d',
                          nocc_b, int(numpy.sum(mo_occ[1])))

            #output 1-dimension occupation number for restricted open-shell
            if mf.istype('ROHF'): mo_occ = mo_occ[0, :] + mo_occ[1, :]
            return mo_occ
    else:
        def get_occ(mo_energy=None, mo_coeff=None):
            if mo_energy is None: mo_energy = mf.mo_energy
            if mo_coeff is None: mo_coeff = mf.mo_coeff
            mo_occ = numpy.zeros_like(setocc)
            nocc = int(setocc.sum())
            s = occorb[:,setocc>0].conj().T.dot(mf.get_ovlp()).dot(mo_coeff)
            #choose a subset of mo_coeff, which maximizes <old|now>
            idx = numpy.argsort(numpy.einsum('ij,ij->j', s, s))[::-1]
            mo_occ[idx[:nocc]] = 1.
            return mo_occ

    mf.get_occ = get_occ
    return mf

mom_occ = mom_occ_

def project_mo_nr2nr(mol1, mo1, mol2):
    r''' Project orbital coefficients from basis set 1 (C1 for mol1) to basis
    set 2 (C2 for mol2).

    .. math::

        |\psi1\rangle = |AO1\rangle C1

        |\psi2\rangle = P |\psi1\rangle = |AO2\rangle S^{-1}\langle AO2| AO1\rangle> C1 = |AO2\rangle> C2

        C2 = S^{-1}\langle AO2|AO1\rangle C1

    There are three relevant functions:
    :func:`project_mo_nr2nr` is the projection for non-relativistic (scalar) basis.
    :func:`project_mo_nr2r` projects from non-relativistic to relativistic basis.
    :func:`project_mo_r2r`  is the projection between relativistic (spinor) basis.
    '''
    s22 = mol2.intor_symmetric('int1e_ovlp')
    s21 = mole.intor_cross('int1e_ovlp', mol2, mol1)
    if isinstance(mo1, numpy.ndarray) and mo1.ndim == 2:
        return lib.cho_solve(s22, numpy.dot(s21, mo1), strict_sym_pos=False)
    else:
        return [lib.cho_solve(s22, numpy.dot(s21, x), strict_sym_pos=False)
                for x in mo1]

@lib.with_doc(project_mo_nr2nr.__doc__)
def project_mo_nr2r(mol1, mo1, mol2):
    assert (not mol1.cart)
    s22 = mol2.intor_symmetric('int1e_ovlp_spinor')
    s21 = mole.intor_cross('int1e_ovlp_sph', mol2, mol1)

    ua, ub = mol2.sph2spinor_coeff()
    s21 = numpy.dot(ua.T.conj(), s21) + numpy.dot(ub.T.conj(), s21) # (*)
    # mo2: alpha, beta have been summed in Eq. (*)
    # so DM = mo2[:,:nocc] * 1 * mo2[:,:nocc].H
    if isinstance(mo1, numpy.ndarray) and mo1.ndim == 2:
        mo2 = numpy.dot(s21, mo1)
        return lib.cho_solve(s22, mo2, strict_sym_pos=False)
    else:
        return [lib.cho_solve(s22, numpy.dot(s21, x), strict_sym_pos=False)
                for x in mo1]

@lib.with_doc(project_mo_nr2nr.__doc__)
def project_mo_r2r(mol1, mo1, mol2):
    s22 = mol2.intor_symmetric('int1e_ovlp_spinor')
    t22 = mol2.intor_symmetric('int1e_spsp_spinor')
    s21 = mole.intor_cross('int1e_ovlp_spinor', mol2, mol1)
    t21 = mole.intor_cross('int1e_spsp_spinor', mol2, mol1)
    n2c = s21.shape[1]
    pl = lib.cho_solve(s22, s21, strict_sym_pos=False)
    ps = lib.cho_solve(t22, t21, strict_sym_pos=False)
    if isinstance(mo1, numpy.ndarray) and mo1.ndim == 2:
        return numpy.vstack((numpy.dot(pl, mo1[:n2c]),
                             numpy.dot(ps, mo1[n2c:])))
    else:
        return [numpy.vstack((numpy.dot(pl, x[:n2c]),
                              numpy.dot(ps, x[n2c:]))) for x in mo1]

def project_dm_nr2nr(mol1, dm1, mol2):
    r''' Project density matrix representation from basis set 1 (mol1) to basis
    set 2 (mol2).

    .. math::

        |AO2\rangle DM_AO2 \langle AO2|

        = |AO2\rangle P DM_AO1 P \langle AO2|

        DM_AO2 = P DM_AO1 P

        P = S_{AO2}^{-1}\langle AO2|AO1\rangle

    There are three relevant functions:
    :func:`project_dm_nr2nr` is the projection for non-relativistic (scalar) basis.
    :func:`project_dm_nr2r` projects from non-relativistic to relativistic basis.
    :func:`project_dm_r2r`  is the projection between relativistic (spinor) basis.
    '''
    s22 = mol2.intor_symmetric('int1e_ovlp')
    s21 = mole.intor_cross('int1e_ovlp', mol2, mol1)
    p21 = lib.cho_solve(s22, s21, strict_sym_pos=False)
    if isinstance(dm1, numpy.ndarray) and dm1.ndim == 2:
        return reduce(numpy.dot, (p21, dm1, p21.conj().T))
    else:
        return lib.einsum('pi,nij,qj->npq', p21, dm1, p21.conj())

@lib.with_doc(project_dm_nr2nr.__doc__)
def project_dm_nr2r(mol1, dm1, mol2):
    assert (not mol1.cart)
    s22 = mol2.intor_symmetric('int1e_ovlp_spinor')
    s21 = mole.intor_cross('int1e_ovlp_sph', mol2, mol1)

    ua, ub = mol2.sph2spinor_coeff()
    s21 = numpy.dot(ua.T.conj(), s21) + numpy.dot(ub.T.conj(), s21) # (*)
    # mo2: alpha, beta have been summed in Eq. (*)
    # so DM = mo2[:,:nocc] * 1 * mo2[:,:nocc].H
    p21 = lib.cho_solve(s22, s21, strict_sym_pos=False)
    if isinstance(dm1, numpy.ndarray) and dm1.ndim == 2:
        return reduce(numpy.dot, (p21, dm1, p21.conj().T))
    else:
        return lib.einsum('pi,nij,qj->npq', p21, dm1, p21.conj())

@lib.with_doc(project_dm_nr2nr.__doc__)
def project_dm_r2r(mol1, dm1, mol2):
    s22 = mol2.intor_symmetric('int1e_ovlp_spinor')
    t22 = mol2.intor_symmetric('int1e_spsp_spinor')
    s21 = mole.intor_cross('int1e_ovlp_spinor', mol2, mol1)
    t21 = mole.intor_cross('int1e_spsp_spinor', mol2, mol1)
    pl = lib.cho_solve(s22, s21, strict_sym_pos=False)
    ps = lib.cho_solve(t22, t21, strict_sym_pos=False)
    p21 = scipy.linalg.block_diag(pl, ps)
    if isinstance(dm1, numpy.ndarray) and dm1.ndim == 2:
        return reduce(numpy.dot, (p21, dm1, p21.conj().T))
    else:
        return lib.einsum('pi,nij,qj->npq', p21, dm1, p21.conj())

def canonical_orth_(S, thr=1e-7):
    '''Löwdin's canonical orthogonalization'''
    # Ensure the basis functions are normalized (symmetry-adapted ones are not!)
    normlz = numpy.power(numpy.diag(S), -0.5)
    Snorm = numpy.dot(numpy.diag(normlz), numpy.dot(S, numpy.diag(normlz)))
    # Form vectors for normalized overlap matrix
    Sval, Svec = numpy.linalg.eigh(Snorm)
    X = Svec[:,Sval>=thr] / numpy.sqrt(Sval[Sval>=thr])
    # Plug normalization back in
    X = numpy.dot(numpy.diag(normlz), X)
    return X

def partial_cholesky_orth_(S, canthr=1e-7, cholthr=1e-9):
    '''Partial Cholesky orthogonalization for curing overcompleteness.

    References:

    Susi Lehtola, Curing basis set overcompleteness with pivoted
    Cholesky decompositions, J. Chem. Phys. 151, 241102 (2019),
    doi:10.1063/1.5139948.

    Susi Lehtola, Accurate reproduction of strongly repulsive
    interatomic potentials, Phys. Rev. A 101, 032504 (2020),
    doi:10.1103/PhysRevA.101.032504.
    '''
    # Ensure the basis functions are normalized
    normlz = numpy.power(numpy.diag(S), -0.5)
    Snorm = numpy.dot(numpy.diag(normlz), numpy.dot(S, numpy.diag(normlz)))

    # Sort the basis functions according to the Gershgorin circle
    # theorem so that the Cholesky routine is well-initialized
    odS = numpy.abs(Snorm)
    numpy.fill_diagonal(odS, 0.0)
    odSs = numpy.sum(odS, axis=0)
    sortidx = numpy.argsort(odSs, kind='stable')

    # Run the pivoted Cholesky decomposition
    Ssort = Snorm[numpy.ix_(sortidx, sortidx)].copy()
    c, piv, r_c = pivoted_cholesky(Ssort, tol=cholthr)
    # The functions we're going to use are given by the pivot as
    idx = sortidx[piv[:r_c]]

    # Get the (un-normalized) sub-basis
    Ssub = S[numpy.ix_(idx, idx)].copy()
    # Orthogonalize sub-basis
    Xsub = canonical_orth_(Ssub, thr=canthr)

    # Full X
    X = numpy.zeros((S.shape[0], Xsub.shape[1]), dtype=Xsub.dtype)
    X[idx,:] = Xsub

    return X

def remove_linear_dep_(mf, threshold=LINEAR_DEP_THRESHOLD,
                       lindep=LINEAR_DEP_TRIGGER,
                       cholesky_threshold=CHOLESKY_THRESHOLD,
                       force_pivoted_cholesky=FORCE_PIVOTED_CHOLESKY):
    '''
    Args:
        threshold : float
            The threshold under which the eigenvalues of the overlap matrix are
            discarded to avoid numerical instability.
        lindep : float
            The threshold that triggers the special treatment of the linear
            dependence issue.
    '''
    s = mf.get_ovlp()
    cond = numpy.max(lib.cond(s))
    if cond < 1./lindep and not force_pivoted_cholesky:
        return mf

    logger.info(mf, 'Applying remove_linear_dep_ on SCF object.')
    logger.debug(mf, 'Overlap condition number %g', cond)
    if (cond < 1./numpy.finfo(s.dtype).eps and not force_pivoted_cholesky):
        logger.info(mf, 'Using canonical orthogonalization with threshold {}'.format(threshold))
        mf._eigh = _eigh_with_canonical_orth(threshold)
    else:
        logger.info(mf, 'Using partial Cholesky orthogonalization '
                    '(doi:10.1063/1.5139948, doi:10.1103/PhysRevA.101.032504)')
        logger.info(mf, 'Using threshold {} for pivoted Cholesky'.format(cholesky_threshold))
        logger.info(mf, 'Using threshold {} to orthogonalize the subbasis'.format(threshold))
        mf._eigh = _eigh_with_pivot_cholesky(threshold, cholesky_threshold)
    return mf
remove_linear_dep = remove_linear_dep_

def _eigh_with_canonical_orth(threshold=LINEAR_DEP_THRESHOLD):
    def eigh(h, s):
        x = canonical_orth_(s, threshold)
        xhx = reduce(lib.dot, (x.conj().T, h, x))
        e, c = scipy.linalg.eigh(xhx)
        c = numpy.dot(x, c)
        return e, c
    return eigh

def _eigh_with_pivot_cholesky(threshold=LINEAR_DEP_THRESHOLD,
                              cholesky_threshold=CHOLESKY_THRESHOLD):
    def eigh(h, s):
        x = partial_cholesky_orth_(s, canthr=threshold, cholthr=cholesky_threshold)
        xhx = reduce(lib.dot, (x.conj().T, h, x))
        e, c = scipy.linalg.eigh(xhx)
        c = numpy.dot(x, c)
        return e, c
    return eigh

def convert_to_uhf(mf, out=None, remove_df=False):
    '''Convert the given mean-field object to the unrestricted HF/KS object

    Note this conversion only changes the class of the mean-field object.
    The total energy and wave-function are the same as them in the input
    mf object. If mf is an second order SCF (SOSCF) object, the SOSCF layer
    will be discarded. Its underlying SCF object mf._scf will be converted.

    Args:
        mf : SCF object

    Kwargs
        remove_df : bool
            Whether to convert the DF-SCF object to the normal SCF object.
            This conversion is not applied by default.

    Returns:
        An unrestricted SCF object
    '''
    from pyscf import scf
    from pyscf import dft
    assert (isinstance(mf, scf.hf.SCF))

    logger.debug(mf, 'Converting %s to UHF', mf.__class__)

    if mf.istype('GHF'):
        raise NotImplementedError

    elif out is not None:
        assert out.istype('UHF')
        out = _update_mf_without_soscf(mf, out, remove_df)

    elif mf.istype('UHF'):
        # Remove with_df for SOSCF method because the post-HF code checks the
        # attribute .with_df to identify whether an SCF object is DF-SCF method.
        # with_df in SOSCF is used in orbital hessian approximation only.  For the
        # returned SCF object, whether with_df exists in SOSCF has no effects on the
        # mean-field energy and other properties.
        if getattr(mf, '_scf', None):
            return _update_mf_without_soscf(mf, mf._scf.copy(), remove_df)
        else:
            return mf.copy()

    else:
        known_cls = {
            dft.rks_symm.ROKS : dft.uks_symm.UKS,
            dft.rks_symm.RKS  : dft.uks_symm.UKS,
            dft.roks.ROKS     : dft.uks.UKS,
            dft.rks.RKS       : dft.uks.UKS,
            scf.hf_symm.ROHF  : scf.uhf_symm.UHF,
            scf.hf_symm.RHF   : scf.uhf_symm.UHF,
            scf.rohf.ROHF     : scf.uhf.UHF,
            scf.hf.RHF        : scf.uhf.UHF,
        }
        out = _object_without_soscf(mf, known_cls, remove_df)

    return _update_mo_to_uhf_(mf, out)

def _object_without_soscf(mf, known_class, remove_df=False):
    '''Create a new SCF object, excluding the SOSCF base class'''
    from pyscf.soscf import newton_ah
    from pyscf.df.df_jk import _DFHF
    if isinstance(mf, newton_ah._CIAH_SOSCF):
        mf = mf.undo_soscf()

    if remove_df and isinstance(mf, _DFHF):
        mf = mf.undo_df()

    for old_cls in mf.__class__.__mro__:
        if old_cls in known_class:
            break
    else:
        raise NotImplementedError(
            "Incompatible object types. Mean-field `mf` class not found in "
            "`known_class` type.\n\nmf = '%s'\n\nknown_class = '%s'" %
            (mf.__class__, known_class))

    new_cls = known_class[old_cls]
    out = new_cls(mf.mol)
    out.__dict__.update(mf.__dict__)
    out.__class__ = lib.replace_class(mf.__class__, old_cls, new_cls)
    return out

def _update_mf_without_soscf(mf, out, remove_df=False):
    '''Update an SCF object, excluding the SOSCF base class'''
    from pyscf.soscf import newton_ah
    mf_dic = dict(mf.__dict__)

    # if mf is SOSCF object, avoid to overwrite the with_df method
    # FIXME: it causes bug when converting pbc-SOSCF.
    if isinstance(mf, newton_ah._CIAH_SOSCF):
        mf_dic.pop('_scf')
        if not hasattr(mf._scf, 'with_df'):
            mf_dic.pop('with_df', None)
    if remove_df:
        mf_dic.pop('with_df', None)

    out.__dict__.update(mf_dic)
    return out

def convert_to_rhf(mf, out=None, remove_df=False):
    '''Convert the given mean-field object to the restricted HF/KS object

    Note this conversion only changes the class of the mean-field object.
    The total energy and wave-function are the same as them in the input
    mf object. If mf is an second order SCF (SOSCF) object, the SOSCF layer
    will be discarded. Its underlying SCF object mf._scf will be converted.

    Args:
        mf : SCF object

    Kwargs
        remove_df : bool
            Whether to convert the DF-SCF object to the normal SCF object.
            This conversion is not applied by default.

    Returns:
        An unrestricted SCF object
    '''
    from pyscf import scf
    from pyscf import dft
    assert (isinstance(mf, scf.hf.SCF))

    logger.debug(mf, 'Converting %s to RHF', mf.__class__)

    if getattr(mf, 'nelec', None) is None:
        nelec = mf.mol.nelec
    else:
        nelec = mf.nelec

    if mf.istype('GHF'):
        raise NotImplementedError

    elif out is not None:
        assert out.istype('RHF')
        out = _update_mf_without_soscf(mf, out, remove_df)

    elif (mf.istype('RHF') or
          (nelec[0] != nelec[1] and mf.istype('ROHF'))):
        if getattr(mf, '_scf', None):
            return _update_mf_without_soscf(mf, mf._scf.copy(), remove_df)
        else:
            return mf.copy()

    else:
        if nelec[0] == nelec[1]:
            known_cls = {
                dft.rks_symm.ROKS: dft.rks_symm.RKS,
                dft.roks.ROKS    : dft.rks.RKS     ,
                dft.uks_symm.UKS : dft.rks_symm.RKS,
                dft.uks.UKS      : dft.rks.RKS     ,
                scf.hf_symm.ROHF : scf.hf_symm.RHF ,
                scf.rohf.ROHF    : scf.hf.RHF      ,
                scf.uhf_symm.UHF : scf.hf_symm.RHF ,
                scf.uhf.UHF      : scf.hf.RHF      ,
            }
        else:
            known_cls = {
                dft.uks_symm.UKS : dft.rks_symm.ROKS,
                dft.uks.UKS      : dft.roks.ROKS    ,
                scf.uhf_symm.UHF : scf.hf_symm.ROHF ,
                scf.uhf.UHF      : scf.rohf.ROHF    ,
            }
        out = _object_without_soscf(mf, known_cls, remove_df)

    return _update_mo_to_rhf_(mf, out)

def convert_to_ghf(mf, out=None, remove_df=False):
    '''Convert the given mean-field object to the generalized HF/KS object

    Note this conversion only changes the class of the mean-field object.
    The total energy and wave-function are the same as them in the input
    mf object. If mf is an second order SCF (SOSCF) object, the SOSCF layer
    will be discarded. Its underlying SCF object mf._scf will be converted.

    Args:
        mf : SCF object

    Kwargs
        remove_df : bool
            Whether to convert the DF-SCF object to the normal SCF object.
            This conversion is not applied by default.

    Returns:
        An generalized SCF object
    '''
    from pyscf import scf
    from pyscf import dft
    assert (isinstance(mf, scf.hf.SCF))

    logger.debug(mf, 'Converting %s to GHF', mf.__class__)

    if out is not None:
        assert out.istype('GHF')
        out = _update_mf_without_soscf(mf, out, remove_df)

    elif mf.istype('GHF'):
        if getattr(mf, '_scf', None):
            return _update_mf_without_soscf(mf, mf._scf.copy(), remove_df)
        else:
            return mf.copy()

    else:
        known_cls = {
            dft.rks_symm.ROKS : dft.gks_symm.GKS,
            dft.rks_symm.RKS  : dft.gks_symm.GKS,
            dft.uks_symm.UKS  : dft.gks_symm.GKS,
            dft.roks.ROKS     : dft.gks.GKS,
            dft.rks.RKS       : dft.gks.GKS,
            dft.uks.UKS       : dft.gks.GKS,
            scf.hf_symm.ROHF  : scf.ghf_symm.GHF,
            scf.hf_symm.RHF   : scf.ghf_symm.GHF,
            scf.uhf_symm.UHF  : scf.ghf_symm.GHF,
            scf.rohf.ROHF     : scf.ghf.GHF,
            scf.hf.RHF        : scf.ghf.GHF,
            scf.uhf.UHF       : scf.ghf.GHF,
        }
        out = _object_without_soscf(mf, known_cls, remove_df)

    return _update_mo_to_ghf_(mf, out)

def _update_mo_to_uhf_(mf, mf1):
    if mf.mo_energy is None:
        return mf1

    if mf.istype('UHF') or mf.istype('KUHF'):
        mf1.mo_occ = mf.mo_occ
        mf1.mo_coeff = mf.mo_coeff
        mf1.mo_energy = mf.mo_energy
    elif getattr(mf, 'kpts', None) is None:  # RHF/ROHF
        mf1.mo_occ = numpy.array((mf.mo_occ>0, mf.mo_occ==2), dtype=numpy.double)
        # ROHF orbital energies, not canonical UHF orbital energies
        mo_ea = getattr(mf.mo_energy, 'mo_ea', mf.mo_energy)
        mo_eb = getattr(mf.mo_energy, 'mo_eb', mf.mo_energy)
        mf1.mo_energy = numpy.array((mo_ea, mo_eb))
        mf1.mo_coeff = numpy.array((mf.mo_coeff, mf.mo_coeff))
    else:  # This to handle KRHF object
        mf1.mo_occ = numpy.array([
            [numpy.asarray(occ> 0, dtype=numpy.double) for occ in mf.mo_occ],
            [numpy.asarray(occ==2, dtype=numpy.double) for occ in mf.mo_occ]])
        mo_ea = getattr(mf.mo_energy, 'mo_ea', mf.mo_energy)
        mo_eb = getattr(mf.mo_energy, 'mo_eb', mf.mo_energy)
        mf1.mo_energy = numpy.array((mo_ea, mo_eb))
        mf1.mo_coeff = numpy.array((mf.mo_coeff, mf.mo_coeff))
    return mf1

def _update_mo_to_rhf_(mf, mf1):
    if mf.mo_energy is None:
        return mf1

    if mf.istype('RHF') or mf.istype('KRHF'): # RHF/ROHF/KRHF/KROHF
        mf1.mo_occ = mf.mo_occ
        mf1.mo_coeff = mf.mo_coeff
        mf1.mo_energy = mf.mo_energy
    elif getattr(mf, 'kpts', None) is None:  # UHF
        mf1.mo_occ = mf.mo_occ[0] + mf.mo_occ[1]
        mf1.mo_energy = mf.mo_energy[0]
        mf1.mo_coeff =  mf.mo_coeff[0]
        if getattr(mf.mo_coeff[0], 'orbsym', None) is not None:
            mf1.mo_coeff = lib.tag_array(mf1.mo_coeff, orbsym=mf.mo_coeff[0].orbsym)
        mf1.converged = False
    else:  # KUHF
        mf1.mo_occ = [occa+occb for occa, occb in zip(*mf.mo_occ)]
        mf1.mo_energy = mf.mo_energy[0]
        mf1.mo_coeff =  mf.mo_coeff[0]
        mf1.converged = False
    return mf1

def _update_mo_to_ghf_(mf, mf1):
    if mf.mo_energy is None:
        return mf1

    if mf.istype('KSCF'):
        raise NotImplementedError('KSCF')
    elif mf.istype('RHF'):
        nao, nmo = mf.mo_coeff.shape
        orbspin = get_ghf_orbspin(mf.mo_energy, mf.mo_occ, True)

        mf1.mo_energy = numpy.empty(nmo*2)
        mf1.mo_energy[orbspin==0] = mf.mo_energy
        mf1.mo_energy[orbspin==1] = mf.mo_energy
        mf1.mo_occ = numpy.empty(nmo*2)
        mf1.mo_occ[orbspin==0] = mf.mo_occ > 0
        mf1.mo_occ[orbspin==1] = mf.mo_occ == 2

        mo_coeff = numpy.zeros((nao*2,nmo*2), dtype=mf.mo_coeff.dtype)
        mo_coeff[:nao,orbspin==0] = mf.mo_coeff
        mo_coeff[nao:,orbspin==1] = mf.mo_coeff
        if getattr(mf.mo_coeff, 'orbsym', None) is not None:
            orbsym = numpy.zeros_like(orbspin)
            orbsym[orbspin==0] = mf.mo_coeff.orbsym
            orbsym[orbspin==1] = mf.mo_coeff.orbsym
            mo_coeff = lib.tag_array(mo_coeff, orbsym=orbsym)
        mf1.mo_coeff = lib.tag_array(mo_coeff, orbspin=orbspin)

    else: # UHF
        nao, nmo = mf.mo_coeff[0].shape
        orbspin = get_ghf_orbspin(mf.mo_energy, mf.mo_occ, False)

        mf1.mo_energy = numpy.empty(nmo*2)
        mf1.mo_energy[orbspin==0] = mf.mo_energy[0]
        mf1.mo_energy[orbspin==1] = mf.mo_energy[1]
        mf1.mo_occ = numpy.empty(nmo*2)
        mf1.mo_occ[orbspin==0] = mf.mo_occ[0]
        mf1.mo_occ[orbspin==1] = mf.mo_occ[1]

        mo_coeff = numpy.zeros((nao*2,nmo*2), dtype=mf.mo_coeff[0].dtype)
        mo_coeff[:nao,orbspin==0] = mf.mo_coeff[0]
        mo_coeff[nao:,orbspin==1] = mf.mo_coeff[1]
        if getattr(mf.mo_coeff[0], 'orbsym', None) is not None:
            orbsym = numpy.zeros_like(orbspin)
            orbsym[orbspin==0] = mf.mo_coeff[0].orbsym
            orbsym[orbspin==1] = mf.mo_coeff[1].orbsym
            mo_coeff = lib.tag_array(mo_coeff, orbsym=orbsym)
        mf1.mo_coeff = lib.tag_array(mo_coeff, orbspin=orbspin)
    return mf1

def get_ghf_orbspin(mo_energy, mo_occ, is_rhf=None):
    '''Spin of each GHF orbital when the GHF orbitals are converted from
    RHF/UHF orbitals

    For RHF orbitals, the orbspin corresponds to first occupied orbitals then
    unoccupied orbitals.  In the occupied orbital space, if degenerated, first
    alpha then beta, last the (open-shell) singly occupied (alpha) orbitals. In
    the unoccupied orbital space, first the (open-shell) unoccupied (beta)
    orbitals if applicable, then alpha and beta orbitals

    For UHF orbitals, the orbspin corresponds to first occupied orbitals then
    unoccupied orbitals.
    '''
    if is_rhf is None:  # guess whether the orbitals are RHF orbitals
        is_rhf = mo_energy[0].ndim == 0

    if is_rhf:
        nmo = mo_energy.size
        nocc = numpy.count_nonzero(mo_occ >0)
        nvir = nmo - nocc
        ndocc = numpy.count_nonzero(mo_occ==2)
        nsocc = nocc - ndocc
        orbspin = numpy.array([0,1]*ndocc + [0]*nsocc + [1]*nsocc + [0,1]*nvir)
    else:
        nmo = mo_energy[0].size
        nocca = numpy.count_nonzero(mo_occ[0]>0)
        nvira = nmo - nocca
        noccb = numpy.count_nonzero(mo_occ[1]>0)
        nvirb = nmo - noccb
        # round(6) to avoid numerical uncertainty in degeneracy
        es = numpy.append(mo_energy[0][mo_occ[0] >0],
                          mo_energy[1][mo_occ[1] >0])
        oidx = numpy.argsort(es.round(6), kind='stable')
        es = numpy.append(mo_energy[0][mo_occ[0]==0],
                          mo_energy[1][mo_occ[1]==0])
        vidx = numpy.argsort(es.round(6), kind='stable')
        orbspin = numpy.append(numpy.array([0]*nocca+[1]*noccb)[oidx],
                               numpy.array([0]*nvira+[1]*nvirb)[vidx])
    return orbspin

del (LINEAR_DEP_THRESHOLD, LINEAR_DEP_TRIGGER)

def fast_newton(mf, mo_coeff=None, mo_occ=None, dm0=None,
                auxbasis=None, dual_basis=None, **newton_kwargs):
    '''This is a wrap function which combines several operations. This
    function first setup the initial guess
    from density fitting calculation then use  for
    Newton solver and call Newton solver.

    Newton solver attributes [max_cycle_inner, max_stepsize, ah_start_tol,
    ah_conv_tol, ah_grad_trust_region, ...] can be passed through ``**newton_kwargs``.
    '''
    from pyscf.lib import logger
    from pyscf import df
    from pyscf.soscf import newton_ah
    if auxbasis is None:
        auxbasis = df.addons.aug_etb_for_dfbasis(mf.mol, 'ahlrichs', beta=2.5)
    if dual_basis:
        mf1 = mf.newton()
        pmol = mf1.mol = newton_ah.project_mol(mf.mol, dual_basis)
        mf1 = mf1.density_fit(auxbasis)
    else:
        mf1 = mf.newton().density_fit(auxbasis)
    mf1.with_df._compatible_format = False
    mf1.direct_scf_tol = 1e-7

    if getattr(mf, 'grids', None):
        from pyscf.dft import gen_grid
        approx_grids = gen_grid.Grids(mf.mol)
        approx_grids.verbose = 0
        approx_grids.level = max(0, mf.grids.level-3)
        mf1.grids = approx_grids

        approx_numint = mf._numint.copy()
        mf1._numint = approx_numint
    for key in newton_kwargs:
        setattr(mf1, key, newton_kwargs[key])

    if mo_coeff is None or mo_occ is None:
        mo_coeff, mo_occ = mf.mo_coeff, mf.mo_occ

    if dm0 is not None:
        mo_coeff, mo_occ = mf1.from_dm(dm0)
    elif mo_coeff is None or mo_occ is None:
        logger.note(mf, '========================================================')
        logger.note(mf, 'Generating initial guess with DIIS-SCF for newton solver')
        logger.note(mf, '========================================================')
        if dual_basis:
            mf0 = mf.copy()
            mf0.mol = pmol
            mf0 = mf0.density_fit(auxbasis)
        else:
            mf0 = mf.density_fit(auxbasis)
        mf0.direct_scf_tol = 1e-7
        mf0.conv_tol = 3.
        mf0.conv_tol_grad = 1.
        if mf0.level_shift == 0:
            mf0.level_shift = .2
        if getattr(mf, 'grids', None):
            mf0.grids = approx_grids
            mf0._numint = approx_numint
# Note: by setting small_rho_cutoff, dft.get_veff function may overwrite
# approx_grids and approx_numint.  It will further changes the corresponding
# mf1 grids and _numint.  If initial guess dm0 or mo_coeff/mo_occ were given,
# dft.get_veff are not executed so that more grid points may be found in
# approx_grids.
            mf0.small_rho_cutoff = mf.small_rho_cutoff * 10
        mf0.kernel()
        mf1.with_df = mf0.with_df
        mo_coeff, mo_occ = mf0.mo_coeff, mf0.mo_occ

        if dual_basis:
            if mo_occ.ndim == 2:
                mo_coeff =(project_mo_nr2nr(pmol, mo_coeff[0], mf.mol),
                           project_mo_nr2nr(pmol, mo_coeff[1], mf.mol))
            else:
                mo_coeff = project_mo_nr2nr(pmol, mo_coeff, mf.mol)
            mo_coeff, mo_occ = mf1.from_dm(mf.make_rdm1(mo_coeff,mo_occ))
        mf0 = None

        logger.note(mf, '============================')
        logger.note(mf, 'Generating initial guess end')
        logger.note(mf, '============================')

    mf1.kernel(mo_coeff, mo_occ)
    mf.converged = mf1.converged
    mf.e_tot     = mf1.e_tot
    mf.mo_energy = mf1.mo_energy
    mf.mo_coeff  = mf1.mo_coeff
    mf.mo_occ    = mf1.mo_occ

#    mf = copy.copy(mf)
#    def mf_kernel(*args, **kwargs):
#        logger.warn(mf, "fast_newton is a wrap function to quickly setup and call Newton solver. "
#                    "There's no need to call kernel function again for fast_newton.")
#        del (mf.kernel)  # warn once and remove circular dependence
#        return mf.e_tot
#    mf.kernel = mf_kernel
    return mf
